from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Generic, Optional, Type, TypeVar, cast

from xcnt.sqlalchemy.search.api_gateway_crd.argument_collection_generator import collection_name_for
from xcnt.sqlalchemy.search.api_gateway_crd.constants import (
    API_VERSION,
    APIGATEWAY_ENTITY_LABEL_KEY,
    APIGATEWAY_ENTITY_META_KEY,
    APIGATEWAY_ROLE_LABEL_KEY,
    APIGATEWAY_ROLE_META_KEY,
    ARGUMENT_COLLECTION_KIND,
)
from xcnt.sqlalchemy.search.api_gateway_crd.fields.api_gateway_converter_plugin import APIGatewayConverterPlugin
from xcnt.sqlalchemy.search.api_gateway_crd.fields.util import api_gateway_field_query
from xcnt.sqlalchemy.search.api_gateway_crd.meta_resolver import find_target_location_for_arguments
from xcnt.sqlalchemy.search.fields import BaseField
from xcnt.sqlalchemy.search.meta_resolver import find_optional_meta_key
from xcnt.sqlalchemy.search.query.resource_base import ResourceBase

if TYPE_CHECKING:
    pass


ToRenderResource = TypeVar("ToRenderResource", bound=ResourceBase)


def default_labels() -> Dict[str, Any]:
    return {"drivr.cloud/generated-for-search": "true"}


def new_template(name: str) -> Dict[str, Any]:
    return {
        "apiVersion": API_VERSION,
        "kind": ARGUMENT_COLLECTION_KIND,
        "metadata": {
            "name": name,
            "labels": default_labels(),
        },
        "spec": {"arguments": {}},
    }


def set_label(name: str, value: str, collection: Dict[str, Any]) -> None:
    labels = collection["metadata"].setdefault("labels", {})
    labels[name] = value


class BaseResourceConverter(APIGatewayConverterPlugin[Type[ToRenderResource]], Generic[ToRenderResource]):
    def to_api_gateway_part(self) -> Optional[Dict[str, Any]]:
        collection = new_template(self.collection_name)
        self._add_labels(collection)

        collection_arguments = collection["spec"]["arguments"]
        parameter_location = self.parameter_location

        fields = cast(Dict[str, BaseField], self.item._fields)
        for key, field in fields.items():
            generated = api_gateway_field_query.api_spec_for(field, self.generator)
            if generated is not None:
                generated["in"] = parameter_location
                collection_arguments[key] = generated

        return collection

    def _add_labels(self, collection: Dict[str, Any]) -> None:
        label = partial(set_label, collection=collection)

        if self.entity_annotation is not None:
            label(APIGATEWAY_ENTITY_LABEL_KEY, self.entity_annotation)
        if self.role_annotation is not None:
            label(APIGATEWAY_ROLE_LABEL_KEY, self.role_annotation)

    @property
    def parameter_location(self) -> str:
        return find_target_location_for_arguments(self.item)

    @property
    def collection_name(self) -> str:
        return collection_name_for(self.item, generator=self.generator)

    @property
    def entity_annotation(self) -> Optional[str]:
        return self.find_optional_meta_key(APIGATEWAY_ENTITY_META_KEY)

    @property
    def role_annotation(self) -> Optional[str]:
        return self.find_optional_meta_key(APIGATEWAY_ROLE_META_KEY)

    def find_optional_meta_key(self, key: str) -> Optional[str]:
        return find_optional_meta_key(self.item, key)
