import inspect
from abc import ABC
from dataclasses import dataclass
from typing import Any, Dict, Generic, Optional, Type, TypeVar, Union, cast

from xcnt.sqlalchemy.search.api_gateway_crd import api_gateway_type_mixin
from xcnt.sqlalchemy.search.api_gateway_crd.argument_collection_generator import ArgumentCollectionRepresentation
from xcnt.sqlalchemy.search.api_gateway_crd.base_resource_converter import new_template
from xcnt.sqlalchemy.search.api_gateway_crd.constants import APIGATEWAY_FILTER_FIELD_KEY, APIGATEWAY_FILTER_TYPE_KEY
from xcnt.sqlalchemy.search.api_gateway_crd.fields.api_gateway_converter_plugin import APIGatewayConverterPlugin
from xcnt.sqlalchemy.search.api_gateway_crd.fields.scalar.filter import FilterRequest
from xcnt.sqlalchemy.search.api_gateway_crd.fields.scalar.filter.converter import FilterConverter
from xcnt.sqlalchemy.search.api_gateway_crd.fields.util import api_spec_for
from xcnt.sqlalchemy.search.documentation_generator import ScalarFieldGenerator
from xcnt.sqlalchemy.search.fields.scalar.base_scalar import BaseScalarField

ScalarField = TypeVar("ScalarField", bound=BaseScalarField)


@dataclass
class ArgumentCollectionRequest(Generic[ScalarField]):
    field: ScalarField

    @property
    def field_class(self) -> Type[ScalarField]:
        return self.field.__class__


class ScalarArgumentCollectionConverter(
    api_gateway_type_mixin.Base,
    APIGatewayConverterPlugin[ArgumentCollectionRequest[ScalarField]],
    Generic[ScalarField],
    ScalarFieldGenerator,
    ABC,
):
    field_class: Type[ScalarField]

    @classmethod
    def supports(self, entity: Any) -> bool:
        if isinstance(entity, ArgumentCollectionRequest):
            return isinstance(entity.field, self.field_class)
        return False

    @property
    def converted_object(self) -> Type[Union[ScalarField, Any]]:
        if inspect.isclass(self.item.field):
            return cast(Type[ScalarField], self.item.field)
        return self.item.field_class

    @property
    def argument_collection_name(self) -> str:
        return self.generator.argument_collection_name_for(
            self.converted_object, name=f"{self.scalar_type_name}-query-field"
        )

    @property
    def current_collection_key(self) -> Any:
        return self.field_class

    @property
    def current_collection(self) -> Optional[ArgumentCollectionRepresentation]:
        return self.generator.collection_for(self.current_collection_key)

    def to_api_gateway_part(self) -> Optional[Dict[str, Any]]:
        existing_collection = self.current_collection
        if existing_collection is not None:
            return existing_collection.spec

        api_gateway_part = self._create_api_gateway_part()
        self.generator.add_argument_collection(
            ArgumentCollectionRepresentation(
                name=self.argument_collection_name, resource=self.current_collection_key, spec=api_gateway_part
            )
        )
        return api_gateway_part

    def _create_api_gateway_part(self) -> Dict[str, Any]:
        argument_collection = new_template(self.argument_collection_name)
        labels = argument_collection["metadata"].setdefault("labels", {})
        labels[APIGATEWAY_FILTER_FIELD_KEY] = self.scalar_type_name
        labels[APIGATEWAY_FILTER_TYPE_KEY] = self.scalar_type_key

        arguments = argument_collection["spec"]["arguments"]

        for filter_name in self.supported_filters:
            arguments[filter_name] = self._filter_reference(filter_name)

        arguments["_null"] = self._null_comparison_filter_reference()
        return argument_collection

    def _null_comparison_filter_reference(self) -> Dict[str, Any]:
        request = FilterRequest(type_name="bool", filter_name="_null", type_key="scalar")
        converter = FilterConverter(item=request, generator=self.generator)
        filter_rule = converter.to_api_gateway_part()
        assert filter_rule is not None
        return filter_rule

    def _filter_reference(self, filter_name: str) -> Dict[str, Any]:
        request = FilterRequest(type_name=self.scalar_type_name, filter_name=filter_name, type_key=self.scalar_type_key)
        filter_rule = api_spec_for(request, self.generator)
        assert filter_rule is not None
        return filter_rule
