import inspect
from enum import Enum
from typing import Any, Dict, Iterable, Optional, cast

from xcnt.sqlalchemy.search.api_gateway_crd.constants import API_VERSION, APIGATEWAY_NAME_KEY, ENUM_KIND
from xcnt.sqlalchemy.search.api_gateway_crd.enum_representation import EnumRepresentation
from xcnt.sqlalchemy.search.api_gateway_crd.fields.api_gateway_converter_plugin import APIGatewayConverterPlugin
from xcnt.sqlalchemy.search.meta_resolver import find_optional_meta_key


def enum_template(enum_name: str) -> Dict[str, Any]:
    return {"apiVersion": API_VERSION, "kind": ENUM_KIND, "metadata": {"name": enum_name}, "spec": {"values": {}}}


class EnumObjectConverter(APIGatewayConverterPlugin[Enum]):
    @classmethod
    def supports(cls, entity: Any) -> bool:
        return inspect.isclass(entity) and issubclass(entity, Enum)

    @property
    def enum_name(self) -> str:
        item = self.item
        assert isinstance(item, type)
        enum_class_name = item.__name__
        api_gateway_name = find_optional_meta_key(item, APIGATEWAY_NAME_KEY)
        if api_gateway_name is None:
            api_gateway_name = enum_class_name

        return self.generator.enum_name_for(self.item, api_gateway_name)

    def to_api_gateway_part(self) -> Optional[Dict[str, Any]]:
        enum_definition = self.generator.enum_for(self.item)
        if enum_definition is not None:
            return enum_definition.spec

        spec = self._generate()
        self.generator.add_enum(
            EnumRepresentation(
                name=self.enum_name,
                spec=spec,
                resource=self.item,
            )
        )
        return spec

    def _generate(self) -> Dict[str, Any]:
        template = enum_template(self.enum_name)
        values = template["spec"]["values"]
        for item in cast(Iterable[Enum], self.item):
            values[item.value] = {}
        return template
