from enum import Enum
from functools import cached_property
from typing import Any, Iterable, List, Optional, Type, Union

from xcnt.sqlalchemy.search.api_gateway_crd.argument_collection_representation import ArgumentCollectionRepresentation
from xcnt.sqlalchemy.search.api_gateway_crd.enum_representation import EnumRepresentation
from xcnt.sqlalchemy.search.api_gateway_crd.resource_generator import ResourceGenerator
from xcnt.sqlalchemy.search.api_gateway_crd.result import Result
from xcnt.sqlalchemy.search.api_gateway_crd.type_holder import TypeHolder
from xcnt.sqlalchemy.search.documentation_generator.resource import resolve_resources
from xcnt.sqlalchemy.search.query.resource_base import ResourceBase


class Generator:
    def __init__(self, definitions: Optional[Iterable[Union[Type[ResourceBase], str]]] = None):
        self._input_definitions = definitions
        self.reset()

    def reset(self) -> None:
        self._argument_collection_holder = TypeHolder[Any, ArgumentCollectionRepresentation]("argument_collection")
        self._enum_holder = TypeHolder[Enum, EnumRepresentation]("enum")

    def add_argument_collection(self, argument_collection: ArgumentCollectionRepresentation) -> None:
        self._argument_collection_holder.add(argument_collection)

    def collection_for(self, resource: Any) -> Optional[ArgumentCollectionRepresentation]:
        return self._argument_collection_holder.mapping_for(resource)

    def argument_collection_name_for(self, resource: Any, name: Optional[str] = None) -> str:
        return self._argument_collection_holder.name_for(resource, name)

    def add_enum(self, enum: EnumRepresentation) -> None:
        self._enum_holder.add(enum)

    def enum_for(self, resource: Enum) -> Optional[EnumRepresentation]:
        return self._enum_holder.mapping_for(resource)

    def enum_name_for(self, resource: Enum, name: Optional[str] = None) -> str:
        return self._enum_holder.name_for(resource, name)

    @cached_property
    def definitions(self) -> List[Type[ResourceBase]]:
        return list(resolve_resources(self._input_definitions))

    def generate(self) -> Result:
        self._generate()
        return Result(
            argument_collections=self._argument_collection_holder.definitions,
            enums=self._enum_holder.definitions,
        )

    def _generate(self) -> None:
        self.reset()

        for definition in self.definitions:
            generator = ResourceGenerator(definition, self)
            generator.generate_if_not_present()
