from __future__ import annotations

from functools import cached_property
from typing import Any, Dict, Tuple, Type, Union

from xcnt.sqlalchemy.search.api_gateway_crd.abstract_reference_converter import AbstractReferenceConverter
from xcnt.sqlalchemy.search.api_gateway_crd.argument_collection_generator import ArgumentCollectionGenerator
from xcnt.sqlalchemy.search.api_gateway_crd.argument_collection_representation import ArgumentCollectionRepresentation
from xcnt.sqlalchemy.search.api_gateway_crd.base_resource_converter import new_template
from xcnt.sqlalchemy.search.api_gateway_crd.fields.util import api_gateway_field_query
from xcnt.sqlalchemy.search.api_gateway_crd.having_reference import HavingReference
from xcnt.sqlalchemy.search.api_gateway_crd.meta_resolver import find_target_location_for_arguments
from xcnt.sqlalchemy.search.api_gateway_crd.mixin import ResourceFieldsMixin
from xcnt.sqlalchemy.search.api_gateway_crd.nested_resource import NestedResource
from xcnt.sqlalchemy.search.parser.having import HAVING_KEY
from xcnt.sqlalchemy.search.query import PartialResource, Resource
from xcnt.sqlalchemy.search.query.resource_having import AbstractHavingBaseClass


class NestedResourceReferenceConverter(
    AbstractReferenceConverter[NestedResource, Union[Resource, PartialResource]], ResourceFieldsMixin
):
    @classmethod
    def supports(self, entity: Any) -> bool:
        return isinstance(entity, NestedResource)

    @property
    def target(self) -> Type[Union[Resource, PartialResource]]:
        return self.item.resource

    @property
    def having(self) -> Type[AbstractHavingBaseClass]:
        return self.target.having

    @property
    def base_argument_collection_name(self) -> str:
        return self.generator.argument_collection_name_for(self.target)

    def _generate_target_argument_collection(self) -> ArgumentCollectionRepresentation:
        self._generate_base_resource_if_necessary()
        self._generate_nested_if_necessary()

        collection = self.generator.collection_for(self.identifier)
        assert collection is not None
        return collection

    def _generate_base_resource_if_necessary(self) -> None:
        resource = self.generator.collection_for(self.target)
        if resource is None:
            self._generate_resource()

    def _generate_resource(self) -> Dict[str, Any]:
        argument_collection_representation = ArgumentCollectionGenerator(self.target, self.generator).generate()
        return argument_collection_representation.spec

    @property
    def identifier(self) -> Tuple[Type[NestedResourceReferenceConverter], Type[Union[Resource, PartialResource]]]:
        return (NestedResourceReferenceConverter, self.target)

    def _generate_nested_if_necessary(self) -> Dict[str, Any]:
        resource = self.generator.collection_for(self.identifier)
        if resource is not None:
            return resource.spec

        return self._generate_nested_resource()

    def _generate_nested_resource(self) -> Dict[str, Any]:
        argument_collection_name = self.collection_name

        base_template = new_template(argument_collection_name)
        representation = ArgumentCollectionRepresentation(
            name=base_template["metadata"]["name"],
            resource=self.identifier,
            spec=base_template,
        )
        self.generator.add_argument_collection(representation)
        base_template["spec"].setdefault("argumentCollections", []).append(self.base_argument_collection_name)
        base_template["spec"]["arguments"][HAVING_KEY] = self._generate_having()
        self.add_self_nested_fields(base_template["spec"]["arguments"])
        return base_template

    def _generate_having(self) -> Dict[str, Any]:
        api_spec = api_gateway_field_query.api_spec_for(HavingReference(self.having), self.generator)
        assert api_spec is not None
        return api_spec

    @cached_property
    def collection_name(self) -> str:  # type: ignore
        return self.generator.argument_collection_name_for(
            self.identifier, "-".join(("nested", self.base_argument_collection_name))
        )

    @property
    def parameter_location(self) -> str:
        return find_target_location_for_arguments(self.target)
