import inspect
from typing import Dict, Generic, List, Optional, Type, TypeVar, Union, cast

import inflection

from xcnt.sqlalchemy.search.api_gateway_crd.yaml_spec import YamlSpec

ObjectType = TypeVar("ObjectType")
YamlHolder = TypeVar("YamlHolder", bound=YamlSpec)


class TypeHolder(Generic[ObjectType, YamlHolder]):
    def __init__(self, type_name: str):
        self._type_name = type_name
        self._names: Dict[str, ObjectType] = {}
        self._resources: Dict[ObjectType, str] = {}
        self._representations: List[YamlHolder] = []
        self._mapping: Dict[ObjectType, YamlHolder] = {}

    @property
    def definitions(self) -> List[YamlHolder]:
        return self._representations

    def mapping_for(self, resource: ObjectType) -> Optional[YamlHolder]:
        return self._mapping.get(resource, None)

    def name_for(self, resource: ObjectType, name: Optional[str] = None) -> str:
        if resource in self._resources:
            return self._resources[resource]

        if name is None:
            name_candidate: Union[Type[ObjectType], ObjectType] = resource
            if not inspect.isclass(name_candidate):
                name = name_candidate.__class__.__name__
            else:
                name = cast(Type[ObjectType], name_candidate).__name__

        name = inflection.dasherize(inflection.underscore(name))
        check_name = name
        index = 0

        while check_name in self._names:
            index += 1
            check_name = f"{check_name}-{index}"

        assert check_name is not None
        self.register_name(name=check_name, resource=resource)

        return self._resources[resource]

    def add(self, representation: YamlHolder) -> None:
        resource = representation.resource
        if resource in self._mapping:
            raise ValueError(f"Resource {resource} is already registered as an {self._type_name}")
        if resource not in self._resources:
            resource_name = representation.name or getattr(resource, "name", None)
            if resource_name is not None:
                self.register_name(resource_name, resource)
        self._mapping[resource] = representation
        self._representations.append(representation)

    def register_name(self, name: str, resource: ObjectType) -> None:
        self._names[name] = resource
        self._resources[resource] = name
