from typing import Any, Callable, Dict, Optional, Type, Union

from apispec import APISpec
from apispec.utils import OpenAPIVersion, build_reference

from xcnt.sqlalchemy.search import registry
from xcnt.sqlalchemy.search.apispec.fields import get_for as get_converter_for
from xcnt.sqlalchemy.search.documentation_generator.self_reference_key import SelfReferenceKey
from xcnt.sqlalchemy.search.documentation_generator.self_reference_key import (
    reference_description as self_reference_description,
)
from xcnt.sqlalchemy.search.documentation_generator.uuid_reference_key import UUIDReferenceKey
from xcnt.sqlalchemy.search.documentation_generator.uuid_reference_key import (
    reference_description as uuid_reference_description,
)
from xcnt.sqlalchemy.search.exceptions import RegistryError
from xcnt.sqlalchemy.search.query import PartialResource, Resource
from xcnt.sqlalchemy.search.query.resource_base import ResourceBase
from xcnt.sqlalchemy.search.query.resource_having import AbstractHavingBaseClass

ResourceType = Type[ResourceBase]


class OpenAPIConverter:
    def __init__(self, openapi_version: Union[str, OpenAPIVersion], spec: APISpec):
        self.openapi_version = OpenAPIVersion(openapi_version)
        self.spec = spec
        self.resolved_schemas: Dict[ResourceType, Dict[str, Any]] = {}
        self.resolved_schema_names: Dict[ResourceType, str] = {}

    def build_reference(self, component_type: str, component_name: str) -> Dict[str, Any]:
        return build_reference(component_type, self.openapi_version.major, component_name)

    def build_schema_reference(self, schema_name: str) -> Dict[str, Any]:
        return self.build_reference("schema", schema_name)

    def resolve_nested_field(self, schema: str) -> Dict[str, Any]:
        self.resolve_nested_schema_reference(schema)
        return self.build_schema_reference(f"Nested{schema}")

    def resolve_reference_field(self, schema: str) -> Dict[str, Any]:
        self.resolve_schema_reference(schema)
        return self.build_schema_reference(schema)

    def resolve_optional_nested_schema_reference(self, schema: str) -> Union[str, Dict[str, Any]]:
        nested_schema = self.resolve_nested_schema(schema)
        if isinstance(nested_schema, str):
            return nested_schema
        else:
            return self.build_schema_reference(schema)

    def resolve_nested_schema_reference(self, schema: str) -> Dict[str, Any]:
        nested_schema_reference = self.resolve_optional_nested_schema_reference(schema)
        if not isinstance(nested_schema_reference, dict):
            raise ValueError(f"Try to dereference nested schema with name {schema} and failing to find it")
        return nested_schema_reference

    def resolve_optional_schema_reference(self, schema: str) -> Union[str, Dict[str, Any]]:
        nested_schema = self.resolve_schema(schema)
        if isinstance(nested_schema, str):
            return nested_schema
        else:
            return self.build_schema_reference(schema)

    def resolve_schema_reference(self, schema: str) -> Dict[str, Any]:
        nested_schema_reference = self.resolve_optional_schema_reference(schema)
        if not isinstance(nested_schema_reference, dict):
            raise ValueError(f"Try to dereference schema with name {schema} and failing to find it")
        return nested_schema_reference

    def resolve_nested_schema(self, schema: str) -> Union[Dict[str, Any], str]:
        resolution = self.check_and_build(schema, self.build_resource_with_nested_resource)
        if resolution is None:
            return schema
        return resolution

    def resolve_schema(self, schema: str) -> Union[Dict[str, Any], str]:
        resolution = self.check_and_build(schema, self.build_resource)
        if resolution is None:
            return schema
        return resolution

    def check_and_build(self, schema: str, builder: Callable[[str, ResourceType], Any]) -> Optional[Dict[str, Any]]:
        try:
            resource = registry.get_class(schema)
        except RegistryError:
            # Do not touch it. Might be a marshmallow schema reference.
            return None

        if resource not in self.resolved_schemas:
            builder(schema, resource)

        return self.resolved_schemas[resource]

    def build_resource_with_nested_resource(self, schema: str, resource: ResourceType) -> Dict[str, Any]:
        self.build_resource(schema, resource)
        return self.build_nested_resource(schema, resource)

    def build_nested_resource(self, name: str, resource: ResourceType) -> Dict[str, Any]:
        schema = self.build_resource(f"Nested{name}", resource)
        having_schema = self._resolve_having_schema(resource)
        if having_schema is not None:
            having_resource = getattr(resource, "having")
            schema["properties"]["_having"] = self.build_schema_reference(self.resolved_schema_names[having_resource])
        return schema

    def build_resource(self, name: str, resource: ResourceType) -> Dict[str, Any]:
        schema: Dict[str, Any] = {"type": "object", "properties": {}}

        if issubclass(resource, (Resource, PartialResource)):
            schema["properties"].update(self._build_reference_properties(name))

        # Pre registering the schema for nested references
        if resource not in self.resolved_schemas:
            self.resolved_schemas[resource] = schema
            self.resolved_schema_names[resource] = name

        having_schema = self._resolve_having_schema_if_necessary(resource)

        if having_schema is not None:
            having_resource = getattr(resource, "having")
            schema["properties"]["_having"] = self.build_schema_reference(self.resolved_schema_names[having_resource])

        resource_fields = getattr(resource, "_fields")
        for key, field in resource_fields.items():
            field_name = field.query_field_name
            if field_name is not None:
                key = field_name
            schema["properties"][key] = self.build_field(field)

        self.spec.components.schema(name, schema)
        return schema

    def _build_reference_properties(
        self,
        name: str,
    ) -> Dict[str, Any]:
        properties: Dict[str, Any] = {}

        for self_reference_key in SelfReferenceKey:
            properties[self_reference_key.value] = {
                "type": "array",
                "description": self_reference_description(self_reference_key),
                "items": self.build_schema_reference(name),
            }

        for uuid_reference_key in UUIDReferenceKey:
            properties[uuid_reference_key.value] = {
                "type": "array",
                "description": uuid_reference_description(uuid_reference_key),
                "items": {"type": "string", "format": "uuid"},
            }

        return properties

    def _resolve_having_schema_if_necessary(self, resource: ResourceType) -> Optional[Dict[str, Any]]:
        if not getattr(resource, "_aggregate", False):
            return None

        return self._resolve_having_schema(resource)

    def _resolve_having_schema(self, resource: ResourceType) -> Optional[Dict[str, Any]]:
        if issubclass(resource, Resource):
            having_resource = resource.having(resource)
        elif issubclass(resource, PartialResource):
            having_resource = resource.having()
        else:
            return None

        if having_resource.__class__ not in self.resolved_schemas:
            self._build_having_resource(having_resource)
        return self.resolved_schemas[having_resource.__class__]

    def _build_having_resource(self, having_resource: AbstractHavingBaseClass) -> Dict[str, Any]:
        having_schema: Dict[str, Any] = {"type": "object", "properties": {}}
        classname = having_resource.__class__.__name__

        self.spec.components.schema(classname, having_schema)
        self.resolved_schemas[having_resource.__class__] = having_schema
        self.resolved_schema_names[having_resource.__class__] = classname
        return having_schema

    def build_field(self, item: Any) -> Dict[str, Any]:
        converter_class = get_converter_for(item)
        converter = converter_class(item=item, open_api_converter=self)
        return converter.to_openapi()
