from __future__ import annotations

import importlib
from typing import Any, Generic, Optional, Type, TypeVar, Union

from xcnt.sqlalchemy.search.fields.base import BaseField

NestedObject = TypeVar("NestedObject")


class AbstractNested(BaseField, Generic[NestedObject]):
    _nested_resource: Optional[Type[NestedObject]] = None
    owner: Optional[Any]

    def __init__(
        self,
        resource: Union[Type[NestedObject], str],
        attribute_name: Optional[str] = None,
        description: Optional[str] = None,
    ):
        self.maybe_nested_resource = resource
        super().__init__(attribute_name=attribute_name, description=description)

    def __get__(self, instance: Optional[Any], owner: Type[Any]) -> AbstractNested:
        self.owner = owner
        return self

    @property
    def nested_resource(self) -> Type[NestedObject]:
        if self._nested_resource is None:
            owner_module: Optional[str] = None
            if self.owner is not None:
                owner_module = self.owner.__module__

            if isinstance(self.maybe_nested_resource, str):
                if "." in self.maybe_nested_resource:
                    module, resource_class = self.maybe_nested_resource.rsplit(".", 1)
                else:
                    module = owner_module or self.__class__.__module__
                    resource_class = self.maybe_nested_resource

                nested_resource: Optional[Type[NestedObject]]
                try:
                    nested_resource = getattr(importlib.import_module(module, owner_module), resource_class, None)
                except ImportError:
                    nested_resource = None

                if nested_resource is None:
                    raise ValueError(
                        f"Could not import a nested resource in import path {self.maybe_nested_resource} "
                        f"from module {owner_module}"
                    )
                self._nested_resource = nested_resource
            else:
                self._nested_resource = self.maybe_nested_resource
        return self._nested_resource

    @nested_resource.setter
    def nested_resource(self, nested_resource: Type[NestedObject]) -> None:
        self._nested_resource = nested_resource
