from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, List, Mapping

from sqlalchemy import not_
from sqlalchemy.sql.elements import ClauseElement, ColumnClause

from xcnt.sqlalchemy.search.parser.abstract_query import AbstractQuery
from xcnt.sqlalchemy.search.parser.evaluation_result import EvaluationErrorCode, EvaluationResult, ResultCombination
from xcnt.sqlalchemy.search.parser.flow_exception import ReturnAll, ReturnNothing
from xcnt.sqlalchemy.search.parser.having import HAVING_KEY, HavingParser
from xcnt.sqlalchemy.search.parser.resource import ResourceParser

if TYPE_CHECKING:
    from xcnt.sqlalchemy.search.query import Resource


class BaseQueryParser(ResourceParser[Mapping[str, Any]], AbstractQuery):
    def __init__(self, query: Mapping[str, Any], resource: Resource):
        self.query = query
        super().__init__(resource)

    def parse(self, parse_context: Dict[str, Any]) -> EvaluationResult:
        self.parse_context = parse_context
        try:
            self.parse_sub_results()
        except ReturnAll:
            pass

        self.parse_evaluations()

        if not self.has_data:
            raise ReturnAll()

        return self._result_from_state()

    def parse_sub_results(self) -> None:
        try:
            self._parse_having_if_necessary()
        except ReturnAll:
            pass

        self._parse_or()

        try:
            self._parse_and()
        except ReturnAll:
            if len(self.sub_results) == 0:
                # No OR statements existing, propagating return_all exception
                raise

    def _parse_having_if_necessary(self) -> None:
        if not issubclass(self.query.__class__, Mapping):
            self.errors[HAVING_KEY] = EvaluationErrorCode.NOT_A_DICT
        elif self.query.get(HAVING_KEY, None):
            self._parse_having()

    def _parse_having(self) -> None:
        result = HavingParser(self.query[HAVING_KEY], self.having, self.resource).parse(self.parse_context)
        self.expressions.extend(result.expressions)
        self.group_expressions.extend(result.group_expressions)
        self.sub_results.extend(result.children)
        self.errors.update(result.direct_errors)

    def _parse_or(self) -> None:
        self._parse_listing("_or", self._apply_or)

    def _parse_and(self) -> None:
        self._parse_listing("_and", self._apply_and)

    def _parse_listing(self, key: str, on_success: Callable[[], Any]) -> None:
        query = self.query
        if key in query:
            if self.expect_list(key, query[key]):
                on_success()

    def _apply_or(self) -> None:
        self._apply_listing("_or", ResultCombination.OR)

    def _apply_and(self) -> None:
        self._apply_listing("_and", ResultCombination.AND)

    def _apply_listing(self, listing_key: str, result_type: ResultCombination) -> EvaluationResult:
        list_of_items: List[Dict[str, Any]] = self.query[listing_key]
        clauses: List[EvaluationResult] = []
        for item in list_of_items:
            result = self.__class__(item, self.resource)
            try:
                clauses.append(result.parse(self.parse_context))
            except ReturnAll:
                if result_type == ResultCombination.OR:
                    raise
            except ReturnNothing:
                if result_type == ResultCombination.AND:
                    raise

        if len(clauses) == 0:
            if result_type == ResultCombination.OR:
                raise ReturnNothing()
            elif result_type == ResultCombination.AND:
                raise ReturnAll()
            else:
                raise RuntimeError("This must not happen")

        evaluation_result = EvaluationResult(key_name=listing_key, result=result_type, children=clauses)
        self.sub_results.append(evaluation_result)
        return evaluation_result

    def parse_evaluations(self) -> None:
        # Please keep the order like this, as parse_one_of and parse_not_of adds things to
        # expressions, and fields are being added as sub results. Due to exception handling
        # the order is important
        self._parse_one_of_and_not_of()
        self._parse_fields_or_raise_nothing()

    def _parse_one_of_and_not_of(self) -> None:
        self._parse_one_of()
        self._parse_not_of()

    def _parse_one_of(self) -> None:
        try:
            self._parse_special_primary_key("_one_of", lambda column, value: column.in_(value))
        except ReturnAll:
            pass

    def _parse_not_of(self) -> None:
        # Negate the exceptions as this does exactly the opposite to _parse_one_of
        try:
            self._parse_special_primary_key("_not_of", lambda column, value: not_(column.in_(value)))
        except ReturnAll:
            raise ReturnNothing()
        except ReturnNothing:
            # This is okay as it means not applying anything as ReturnNothing = ReturnAll and True = False
            pass

    def _parse_special_primary_key(
        self, key_name: str, on_value: Callable[[ColumnClause, List[Any]], ClauseElement]
    ) -> None:
        try:
            check_key = self._get_existing_key(key_name)
        except KeyError:
            return

        check_value = self.query[check_key]
        if check_value is None:
            raise ReturnAll()
        if not isinstance(check_value, (tuple, list)):
            self.errors[check_key] = EvaluationErrorCode.NOT_A_LIST
            return
        elif isinstance(check_value, tuple):
            check_value = list(check_value)

        if len(check_value) == 0:
            raise ReturnNothing()

        self.expressions.append(on_value(self._primary_key_column, check_value))

    @property
    def _primary_key_column(self) -> ColumnClause:
        return self.resource.primary_key_column

    def _result_from_state(self) -> EvaluationResult:
        evaluation_result = super()._result_from_state()
        evaluation_result.unaliased_primary_key_reference = self._root_model_primary_key_column
        evaluation_result.root_primary_key_reference = self._primary_key_column

        return evaluation_result
