from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional, TypeVar, Union, cast

import inflection
from sqlalchemy import and_, or_
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm.query import Query
from sqlalchemy.sql import select
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy.sql.expression import ClauseElement, Select


class EvaluationErrorCode(str, Enum):  # Inherit from str to make the Enum JSON serializable
    NOT_A_LIST = "not_a_list"
    UNKNOWN_FIELD = "unknown_field"
    NOT_A_DICT = "query_field_not_a_mapping"
    UNKNOWN_FILTER_FUNCTION = "unknown_filter_function"
    UNSUPPORTED_FILTER_FUNCTION = "unsupported_filter_function"
    INVALID_VALUE_FOR_FILTER = "invalid_filter_value"


class ResultCombination(Enum):
    AND = "and"
    OR = "or"


T = TypeVar("T", bound="NestedErrorItem")
NestedErrorItem = Union[EvaluationErrorCode, Dict[Union[str, int], T]]
NestedErrorCodes = Dict[str, NestedErrorItem]


@dataclass
class JoinCondition:
    base_model: DeclarativeMeta
    model: DeclarativeMeta
    clause: ClauseElement
    base_model_primary_key_column_name: str

    def __hash__(self) -> int:
        return (
            self.model.__hash__()
            ^ self.clause.__hash__()  # noqa: W503
            ^ self.base_model.__hash__()  # noqa: W503
            ^ self.base_model_primary_key_column_name.__hash__()  # noqa: W503
        )

    def apply_to(self, query: Query) -> Query:
        return query.outerjoin(self.model, self.clause)

    def apply_to_select(self, select: Select) -> Select:
        return select.outerjoin(self.model, self.clause)


@dataclass
class EvaluationResult:
    key_name: Optional[str] = None
    result: ResultCombination = ResultCombination.AND
    join_condition: Optional[JoinCondition] = None
    to_join: List[JoinCondition] = field(default_factory=list)
    expressions: List[ClauseElement] = field(default_factory=list)
    group_expressions: List[ClauseElement] = field(default_factory=list)
    group_by_clauses: List[ClauseElement] = field(default_factory=list)
    root_group_expressions: List[ClauseElement] = field(default_factory=list)
    root_group_by_clauses: List[ClauseElement] = field(default_factory=list)
    children: List[EvaluationResult] = field(default_factory=list)
    direct_errors: NestedErrorCodes = field(default_factory=dict)
    root_primary_key_reference: Optional[ColumnElement] = None
    unaliased_primary_key_reference: Optional[ColumnElement] = None

    @classmethod
    def empty(cls) -> EvaluationResult:
        return cls()

    def __add__(self, other: EvaluationResult) -> EvaluationResult:
        assert isinstance(other, EvaluationResult)
        assert other.key_name is None, "Cannot add evaluation result with set key name"
        assert other.join_condition is None, "Cannot add evaluation result with configured join condition"

        new_result = self.clone()
        for to_join in other.to_join:
            if to_join not in new_result.to_join:
                new_result.to_join.append(to_join)
        new_result.expressions.extend(other.expressions)
        new_result.group_expressions.extend(other.group_expressions)
        new_result.group_by_clauses.extend(other.group_by_clauses)
        new_result.root_group_expressions.extend(other.root_group_expressions)
        new_result.children.extend(other.children)
        new_result.direct_errors.update(other.direct_errors)
        return new_result

    def clone(self) -> EvaluationResult:
        return self.__class__(
            key_name=self.key_name,
            result=self.result,
            join_condition=self.join_condition,
            to_join=self.to_join,
            expressions=self.expressions,
            group_expressions=self.group_expressions,
            group_by_clauses=self.group_by_clauses,
            root_group_expressions=self.root_group_expressions,
            children=self.children,
            direct_errors=self.direct_errors,
            root_primary_key_reference=self.root_primary_key_reference,
            unaliased_primary_key_reference=self.unaliased_primary_key_reference,
        )

    def build_clauses(self) -> Optional[ClauseElement]:
        clauses: List[ClauseElement] = self.expressions[:]
        having_clauses: List[ClauseElement] = list(self.direct_group_expressions)

        having_select = None
        if len(having_clauses) > 0:
            if self.join_condition is not None:
                having_select = self._having_select()
            else:
                self.root_group_expressions.extend(having_clauses)
                self.root_group_by_clauses.extend(self.direct_group_by_clauses)

        for child in self.children:
            having_select = self._process_child(child, clauses, having_select)

        return self._process_where_clause(clauses, having_select)

    @property
    def has_direct_group_expressions(self) -> bool:
        for _ in self.direct_group_expressions:
            return True
        return False

    @property
    def direct_group_expressions(self) -> Iterable[ClauseElement]:
        group_expressions = list(self._direct_non_result_conform_group_expressions)
        if not len(group_expressions):
            return

        yield self._to_apply_clause(*group_expressions)

    @property
    def _direct_non_result_conform_group_expressions(self) -> Iterable[ClauseElement]:
        yield from self.group_expressions
        for child in self.children:
            if child.join_condition is None:
                yield from child.direct_group_expressions

    @property
    def direct_group_by_clauses(self) -> Iterable[ClauseElement]:
        yield from self.group_by_clauses
        for child in self.children:
            if child.join_condition is None:
                yield from child.group_by_clauses

    @property
    def has_joins(self) -> bool:
        return self.join_condition is not None or len(self.to_join) > 0

    def build_root_having_clause(self) -> Optional[ClauseElement]:
        if self.has_root_group_expressions:
            return and_(*self.root_group_expressions)
        else:
            return None

    def _process_child(
        self, child: EvaluationResult, clauses: List[ClauseElement], having_select: Optional[Select]
    ) -> Optional[Select]:
        child_clause = child.build_clauses()

        if child_clause is not None:
            clauses.append(child_clause)
            if not child.has_direct_group_expressions and child.has_joins:
                if having_select is not None:
                    having_select = self._apply_joins_to_select(having_select, child)
                else:
                    self._mark_as_to_join(child)

        return having_select

    def _apply_joins_to_select(self, having_select: Select, child: EvaluationResult) -> Select:
        if child.join_condition is not None:
            having_select = child.join_condition.apply_to_select(having_select)
        for join_condition in child.to_join:
            having_select = join_condition.apply_to_select(having_select)
        return having_select

    def _process_where_clause(
        self, clauses: List[ClauseElement], having_select: Optional[Select]
    ) -> Optional[ClauseElement]:

        where_clause: Optional[ClauseElement] = None

        if len(clauses) > 0:
            where_clause = self._to_apply_clause(*clauses)

        if having_select is not None:
            if where_clause is not None:
                having_select = having_select.where(where_clause)
            return self._root_model_primary_column.in_(having_select.scalar_subquery())
        else:
            return where_clause

    @property
    def _to_apply_clause(self) -> Callable:
        if self.result == ResultCombination.AND:
            to_apply_clause = and_
        elif self.result == ResultCombination.OR:
            to_apply_clause = or_
        else:
            raise ValueError("Unknown result combination ", self.result)
        return to_apply_clause

    @property
    def _root_model_primary_column(self) -> ColumnElement:
        join_condition = self.join_condition
        if self.join_condition is None:
            return self.root_primary_key_reference

        base_model = getattr(join_condition, "base_model")
        base_model_primary_key_column_name = getattr(join_condition, "base_model_primary_key_column_name")
        root_model_primary_column = getattr(base_model, base_model_primary_key_column_name)

        return root_model_primary_column

    def _sub_select(self) -> Select:
        stmt = select(self._root_model_primary_column)
        return stmt

    def _having_select(self) -> Select:
        stmt = self._sub_select()
        if self.join_condition is not None:
            stmt = self.join_condition.apply_to_select(stmt)

        stmt = stmt.group_by(self._root_model_primary_column, *self.direct_group_by_clauses)

        direct_group_expressions = list(self.direct_group_expressions)
        if len(direct_group_expressions):
            stmt = stmt.having(self._to_apply_clause(*direct_group_expressions))
        return stmt

    def _mark_as_to_join(self, result: EvaluationResult) -> None:
        if result.join_condition is not None:
            if result.join_condition not in self.to_join:
                self.to_join.append(result.join_condition)
        for to_join in result.to_join:
            if to_join not in self.to_join:
                self.to_join.append(to_join)

    def apply_to_query(self, query: Query) -> Query:
        if self.root_primary_key_reference is not None:
            session = query.session
            to_apply_query = session.query(self.root_primary_key_reference)
            to_apply_query = self._apply_filter_to_query(to_apply_query)

            if to_apply_query is not None:
                query = query.filter(self.root_primary_key_reference.in_(to_apply_query.subquery().select()))
        else:
            to_apply_query = self._apply_filter_to_query(query)
            if to_apply_query is not None:
                query = to_apply_query

        having_clause = self.build_root_having_clause()
        if having_clause is not None:
            query = query.having(having_clause)

        direct_group_by_clauses = tuple(self.direct_group_by_clauses)
        if len(direct_group_by_clauses) > 0:
            query = query.group_by(*direct_group_by_clauses)

        return query

    def _apply_filter_to_query(self, query: Query) -> Optional[Query]:
        where_clause = self.build_clauses()

        if where_clause is None:
            return None

        query = self._apply_joins(query)

        if where_clause is not None:
            query = query.filter(where_clause)

        return query

    def _apply_joins(self, query: Query) -> Query:
        for to_join in self.to_join:
            query = to_join.apply_to(query)
        return query

    @property
    def errors(self) -> NestedErrorCodes:
        errors: NestedErrorCodes = self.direct_errors.copy()
        for child_result in self.children:
            key_name = child_result.key_name
            sub_errors = errors
            child_errors = child_result.errors
            if len(child_errors) == 0:
                continue

            if key_name is not None:
                sub_errors = cast(NestedErrorCodes, sub_errors.setdefault(key_name, {}))
            sub_errors.update(child_errors)
        return errors

    @property
    def has_group_expressions(self) -> bool:
        return len(self.group_expressions) > 0

    @property
    def has_root_group_expressions(self) -> bool:
        return len(self.root_group_expressions) > 0


def key_derivations(key: str) -> Iterable[str]:
    transformations: Iterable[Callable[[str], str]] = (
        (lambda value: value),
        partial(inflection.camelize, uppercase_first_letter=False),
        inflection.underscore,
    )
    for transformation in transformations:
        yield transformation(key)


class ErrorResultParser(ABC):
    def __init__(self) -> None:
        self.errors: NestedErrorCodes = {}

    def expect_list(self, key: str, payload: Any) -> bool:
        is_a_list = isinstance(payload, list)
        if not is_a_list:
            self.errors[key] = EvaluationErrorCode.NOT_A_LIST
        return is_a_list


class EvaluationResultParser(ErrorResultParser, ABC):
    def __init__(self) -> None:
        super().__init__()
        self.sub_results: List[EvaluationResult] = []
        self.expressions: List[ClauseElement] = []
        self.group_expressions: List[ClauseElement] = []
        self.group_by_clauses: List[ClauseElement] = []

    @property
    def has_data(self) -> bool:
        return any(
            len(item) > 0
            for item in [
                cast(dict, self.errors),
                self.sub_results,
                self.expressions,
                self.group_expressions,
                self.group_by_clauses,
            ]
        )

    @abstractmethod
    def parse(self, parse_context: Dict[str, Any]) -> EvaluationResult:
        pass

    @property
    def result_combination(self) -> ResultCombination:
        return ResultCombination.AND

    def _result_from_state(self) -> EvaluationResult:
        return EvaluationResult(
            result=self.result_combination,
            expressions=self.expressions,
            group_expressions=self.group_expressions,
            children=self.sub_results,
            direct_errors=self.errors,
            group_by_clauses=self.group_by_clauses,
        )
