from __future__ import annotations

from abc import ABCMeta
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Type, TypeVar

import pkg_resources

from xcnt.sqlalchemy.search.parser.evaluation_result import EvaluationResult

if TYPE_CHECKING:
    from xcnt.sqlalchemy.search.fields.base import BaseField
    from xcnt.sqlalchemy.search.parser.field_plugin import AbstractFieldPlugin


PLUGIN_ENTRY_POINT = "xcnt.sqlalchemy.search.field.parser"


def _supported_fields_from_entrypoints() -> List[Type[AbstractFieldPlugin]]:
    return [entry_point.load() for entry_point in pkg_resources.iter_entry_points(PLUGIN_ENTRY_POINT)]


def add_field_plugin(field_plugin: Type[AbstractFieldPlugin]) -> None:
    FieldParser.supported_fields.append(field_plugin)


class FieldParserMeta(ABCMeta):
    @property
    def supported_fields(self) -> List[Type[AbstractFieldPlugin]]:
        if not hasattr(self, "_cached_supported_fields"):
            self._cached_supported_fields = _supported_fields_from_entrypoints()
        return self._cached_supported_fields


QueryType = TypeVar("QueryType")


class FieldParser(Generic[QueryType], metaclass=FieldParserMeta):
    accepts_non_dict: bool = False

    def __init__(self, key: str, query: QueryType, resource_field: BaseField):
        self.query = query
        self.key = key
        self.resource_field = resource_field

    def parse(self, parse_context: Dict[str, Any]) -> EvaluationResult:
        return self.appropriate_field_parser.parse(parse_context)

    @property
    def appropriate_field_parser(self) -> AbstractFieldPlugin:
        resource_field = self.resource_field
        supported_fields = self.__class__.supported_fields
        for field_plugin_class in supported_fields:
            if field_plugin_class.supports(resource_field):  # type: ignore
                return field_plugin_class(self.key, self.query, resource_field)

        raise ValueError(f"Unknown field for parsing {resource_field}")
