from __future__ import annotations

from abc import ABCMeta, abstractproperty
from distutils.util import strtobool
from typing import Any, Dict, Optional

import pkg_resources
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.sql.elements import ClauseElement, ColumnClause

from xcnt.sqlalchemy.search.fields.scalar.base_scalar import BaseScalarField
from xcnt.sqlalchemy.search.parser.evaluation_result import EvaluationErrorCode, EvaluationResult
from xcnt.sqlalchemy.search.parser.field_plugin import ConcreteField, FieldPlugin
from xcnt.sqlalchemy.search.parser.filters import FilterClass, InvalidSearchData
from xcnt.sqlalchemy.search.parser.flow_exception import ReturnAll

PLUGIN_ENTRY_POINT = "xcnt.sqlalchemy.search.field.filter"
NULL_FILTER = "null"
NULL_VALUE_CHECK_KEY = "_null"


def _function_mapping_from_entrypoints() -> Dict[str, FilterClass]:
    return dict(
        (entry_point.name, entry_point.load()) for entry_point in pkg_resources.iter_entry_points(PLUGIN_ENTRY_POINT)
    )


def add_filter_plugin(filter_name: str, filter_plugin: FilterClass) -> None:
    FilterFieldParser.function_mapping[filter_name] = filter_plugin


class FilterParserMeta(ABCMeta):
    @property
    def function_mapping(self) -> Dict[str, FilterClass]:
        if not hasattr(self, "_cached_function_mapping"):
            self._cached_function_mapping = _function_mapping_from_entrypoints()
        return self._cached_function_mapping


class FilterFieldParser(FieldPlugin[ConcreteField, Dict[str, Any]], metaclass=FilterParserMeta):
    @property
    def _function_mapping(self) -> Dict[str, FilterClass]:
        return self.__class__.function_mapping

    def parse(self, parse_context: Dict[str, Any]) -> EvaluationResult:
        self.parse_context = parse_context
        self.parse_filters()

        return self._result_from_state()

    def parse_filters(self) -> None:
        if self.query is None:
            self.errors[NULL_FILTER] = EvaluationErrorCode.NOT_A_DICT
            return
        filter_names = self.query.keys()
        if len(filter_names) == 1 and NULL_VALUE_CHECK_KEY in filter_names:
            result = self.parse_null_filter()
            self.add_expression(result)
        else:
            for filter_name in filter_names:
                self.parse_filter(filter_name)

    @abstractproperty
    def model_field(self) -> ColumnClause:
        pass

    @property
    def is_model_field_hybrid_property(self) -> bool:
        descriptor = getattr(self.model_field, "descriptor", None)
        return descriptor is not None and isinstance(descriptor, hybrid_property)

    def parse_filter(self, filter_name: str) -> None:
        result = self.filter_for(filter_name)
        if result is not None:
            self.add_expression(result)

    def add_expression(self, expression: ClauseElement) -> None:
        self.expressions.append(expression)

    @property
    def mapping_field(self) -> BaseScalarField:
        resource_field = self.resource_field
        assert isinstance(resource_field, BaseScalarField)
        return resource_field

    def filter_for(self, filter_name: str) -> Optional[ClauseElement]:
        if filter_name == NULL_VALUE_CHECK_KEY:
            return None

        applied_mapping = self._function_mapping.get(filter_name, None)
        if applied_mapping is None:
            self.errors[filter_name] = EvaluationErrorCode.UNKNOWN_FILTER_FUNCTION
            return None

        try:
            mapping = applied_mapping(self.mapping_field, self.model_field)
        except ReturnAll:
            return None

        if not mapping.supported:
            self.errors[filter_name] = EvaluationErrorCode.UNSUPPORTED_FILTER_FUNCTION
            return None

        try:
            return_null_values = self._extract_null_filter_value()
        except InvalidSearchData:
            self.errors[filter_name] = {NULL_VALUE_CHECK_KEY: EvaluationErrorCode.INVALID_VALUE_FOR_FILTER}
            return None

        try:
            return mapping.apply(self.query[filter_name], return_null_values)
        except InvalidSearchData:
            self.errors[filter_name] = EvaluationErrorCode.INVALID_VALUE_FOR_FILTER
            return None
        except ReturnAll:
            null_value_filter = self.query.get(NULL_VALUE_CHECK_KEY, None)
            if null_value_filter is not None and not null_value_filter:
                return mapping.is_not_none_expression
            raise

    def _extract_null_filter_value(self) -> bool:
        value = self.query.get(NULL_VALUE_CHECK_KEY, False)
        if isinstance(value, str):
            try:
                value = strtobool(value)
            except ValueError:
                raise InvalidSearchData("Invalid input for _null")
        if not isinstance(value, bool):
            raise InvalidSearchData("Invalid input for _null")
        return value

    def parse_null_filter(self) -> Optional[ClauseElement]:
        try:
            value = self._extract_null_filter_value()
            if value and self.is_model_field_hybrid_property:
                return self.model_field is None
            elif value and not self.is_model_field_hybrid_property:
                return self.model_field.is_(None)
            elif value is None and self.is_model_field_hybrid_property:
                return self.model_field is not None
            else:
                return self.model_field.isnot(None)
        except InvalidSearchData:
            self.errors[NULL_VALUE_CHECK_KEY] = EvaluationErrorCode.INVALID_VALUE_FOR_FILTER
            return None
