from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any

from sqlalchemy import and_, or_
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.sql.elements import ColumnClause
from sqlalchemy.sql.expression import ClauseElement

from xcnt.sqlalchemy.search.meta_resolver import find_meta_key

if TYPE_CHECKING:
    from xcnt.sqlalchemy.search.fields.scalar.base_scalar import BaseScalarField


class FilterMeta(ABCMeta):
    @property
    def key(self) -> str:
        return find_meta_key(self, "key")

    @property
    def is_listed(self) -> bool:
        return find_meta_key(self, "is_listed")


class BaseFilterParser(ABC, metaclass=FilterMeta):
    class Meta:
        is_listed: bool = False

    def __init__(self, field: BaseScalarField, column: ColumnClause):
        self.field = field
        self.model_field = column

    def apply(self, value: Any, return_null_values: bool = False) -> ClauseElement:
        normalized_value = self.normalize_value(value)
        expression = self._apply_normalized(normalized_value)
        return self._filter_null_values(expression, return_null_values)

    def _filter_null_values(self, expression: ClauseElement, return_null_values: bool) -> ClauseElement:
        if not return_null_values:
            return and_(self.is_not_none_expression, expression)
        return or_(self.is_none_expression, expression)

    @property
    def is_not_none_expression(self) -> ClauseElement:
        if self.is_model_field_hybrid_property:
            return self.model_field is not None
        return self.model_field.isnot(None)

    @property
    def is_none_expression(self) -> ClauseElement:
        if self.is_model_field_hybrid_property:
            return self.model_field is None
        return self.model_field.is_(None)

    @property
    def is_model_field_hybrid_property(self) -> bool:
        descriptor = getattr(self.model_field, "descriptor", None)
        return descriptor is not None and isinstance(descriptor, hybrid_property)

    @property
    def key(self) -> str:
        return self.__class__.key  # type: ignore

    @abstractmethod
    def _apply_normalized(self, value: Any) -> ClauseElement:
        pass

    def normalize_value(self, value: Any) -> Any:
        return self.field.convert(value)

    @property
    def supported(self) -> bool:
        try:
            return self.field.meta_key(self.key)
        except KeyError:
            return False
