from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterable, Iterator, Mapping, Tuple

from xcnt.sqlalchemy.search.parser.abstract_query import AbstractQuery
from xcnt.sqlalchemy.search.parser.computed import ComputedField
from xcnt.sqlalchemy.search.parser.evaluation_result import EvaluationErrorCode, EvaluationResult, key_derivations
from xcnt.sqlalchemy.search.parser.field import FieldParser
from xcnt.sqlalchemy.search.parser.flow_exception import ReturnAll
from xcnt.sqlalchemy.search.parser.having_group import HavingGroup
from xcnt.sqlalchemy.search.parser.resource import ResourceParser

if TYPE_CHECKING:
    from xcnt.sqlalchemy.search.parser.field import BaseField
    from xcnt.sqlalchemy.search.query.resource import Resource
    from xcnt.sqlalchemy.search.query.resource_having import HavingBaseClass


HAVING_KEY = "_having"


class HavingParser(ResourceParser[Mapping[str, Any]], AbstractQuery[Mapping[str, Any]]):
    def __init__(self, query: Mapping[str, Any], having_class: HavingBaseClass, resource: Resource):
        super().__init__(resource)
        self.query = query
        self.having_class = having_class

    def parse(self, parse_context: Dict[str, Any]) -> EvaluationResult:
        self.parse_context = parse_context
        self._parse_having()

        if not self.has_data:
            raise ReturnAll()

        return self._result_from_state()

    def _parse_having(self) -> None:
        self.sub_results.extend(self._parse_having_fields())

    def _parse_having_fields(self) -> Iterable[EvaluationResult]:
        for key, query_config, resource_field in self.having_query_fields:
            if isinstance(resource_field, (ComputedField, HavingGroup)):
                parser = FieldParser(key, query_config, resource_field)
                yield parser.parse(self.parse_context)
            else:
                raise AttributeError(
                    f"Field with key {key} does not point to a ComputedField or HavingGroup field. "
                    f"HavingClasses don't support other. Got: {resource_field}"
                )

    @property
    def having_query_fields(self) -> Iterator[Tuple[str, Dict[str, Any], BaseField]]:
        for key, value in self.query.items():
            try:
                resource_key = self._get_having_key(key)
            except KeyError:
                resource_key = key

            if resource_key not in self.having_fields:
                self.errors[key] = EvaluationErrorCode.UNKNOWN_FIELD
            else:
                resource_field = self.having_fields[resource_key]
                yield key, value, resource_field

    def _get_having_key(self, key: str) -> str:
        for check_key in key_derivations(key):
            if check_key in self.having_fields:
                return check_key
        raise KeyError(f'Having key "{key}" not found')

    @property
    def having_fields(self) -> Dict[str, BaseField]:
        return self.having_class._fields

    def _result_from_state(self) -> EvaluationResult:
        new_sub_results = []

        for sub_result in self.sub_results:
            assert len(sub_result.children) == 0, "Having filters cannot be nested"
            self.group_expressions.extend(sub_result.expressions)
            self.group_expressions.extend(sub_result.group_expressions)
            errors = sub_result.errors
            if errors:
                new_sub_results.append(EvaluationResult(key_name=sub_result.key_name, direct_errors=errors))

        assert len(self.expressions) == 0, "Having filters cannot have expressions"

        return EvaluationResult(
            expressions=self.expressions,
            group_expressions=self.group_expressions,
            children=new_sub_results,
            direct_errors=self.errors,
        )
