from __future__ import annotations

from abc import abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, Any, Dict, Generic, TypeVar

from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import RelationshipProperty, aliased
from sqlalchemy.sql.elements import ClauseElement

from xcnt.sqlalchemy.search.fields import JoinConditionConfig
from xcnt.sqlalchemy.search.parser.evaluation_result import EvaluationResult, JoinCondition
from xcnt.sqlalchemy.search.parser.field_plugin import FieldPlugin

if TYPE_CHECKING:
    pass

JoinConditionField = TypeVar("JoinConditionField", bound=JoinConditionConfig)


class JoinedFieldParser(FieldPlugin[JoinConditionField, Dict[str, Any]], Generic[JoinConditionField]):
    def parse(self, parse_context: Dict[str, Any]) -> EvaluationResult:
        nested_result = self._generate_nested(parse_context)
        return self._combine_results(nested_result)

    def _combine_results(self, nested_result: EvaluationResult) -> EvaluationResult:
        result = self._result_from_state()
        return result + nested_result

    @abstractmethod
    def _generate_nested(self, parse_context: Dict[str, Any]) -> EvaluationResult:
        pass

    @property
    @abstractmethod
    def _non_aliased_model(self) -> DeclarativeMeta:
        pass

    @property
    @abstractmethod
    def base_model(self) -> DeclarativeMeta:
        pass

    @property
    def primary_key_column_name(self) -> str:
        return self.resource.primary_key_column_name

    @cached_property
    def model(self) -> DeclarativeMeta:  # type: ignore
        return aliased(self._non_aliased_model)

    def _result_from_state(self) -> EvaluationResult:
        evaluation_result = super()._result_from_state()
        assert self.resource is not None
        evaluation_result.join_condition = JoinCondition(
            base_model=self.base_model,
            model=self.model,
            clause=self._join_clause,
            base_model_primary_key_column_name=self.primary_key_column_name,
        )
        return evaluation_result

    @property
    def _join_clause(self) -> ClauseElement:
        resource_field = self.resource_field
        if resource_field.infer_join_condition:
            return self._inferred_join_clause
        else:
            return resource_field.run_join_condition(self.base_model, self.model)

    @property
    def _inferred_join_clause(self) -> ClauseElement:
        relationship_name = getattr(self.resource_field, "name", None)
        assert relationship_name is not None
        model_field_name = self.resource_field.model_attribute_name
        if model_field_name is None:
            model_field_name = relationship_name

        try:
            maybe_relation = getattr(self.base_model, model_field_name)
        except AttributeError as err:
            raise AttributeError(
                f"The relationship configuration in {self.resource._name} for {model_field_name} "
                "points to an non referenceable attribute",
                err,
            )

        if isinstance(maybe_relation.property, RelationshipProperty):
            return maybe_relation
        else:
            raise AttributeError(
                f"The relationship configuration in {self.resource._name} for {relationship_name} "
                f"does not point to a relationship. Join condition can not be inferred. Got: {maybe_relation}."
            )
