from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, cast

from sqlalchemy.orm import Query, Session

from xcnt.sqlalchemy.search import fields
from xcnt.sqlalchemy.search.fields.concrete_resolver import ResolverRequest
from xcnt.sqlalchemy.search.fields.scalar.base_scalar import BaseScalarField
from xcnt.sqlalchemy.search.parser.abstract_query import AbstractQuery
from xcnt.sqlalchemy.search.parser.evaluation_result import NestedErrorCodes, NestedErrorItem
from xcnt.sqlalchemy.search.parser.field import FieldParser
from xcnt.sqlalchemy.search.parser.having import HAVING_KEY
from xcnt.sqlalchemy.search.parser.partial.resource import PartialResourceParser

if TYPE_CHECKING:
    from xcnt.sqlalchemy.search.query import PartialResource


@dataclass
class ParsingResult:
    result: Mapping[str, Any]
    errors: NestedErrorCodes


class QueryParser(PartialResourceParser, AbstractQuery):
    context: Dict[str, Any]
    partial_result: Dict[str, Any]
    session: Session

    def __init__(self, resource: PartialResource, query: Dict[str, Any]):
        super().__init__(resource)
        self.query = query

    def parse(self, context: Optional[Dict[str, Any]] = None) -> ParsingResult:
        context = self.context = context or {}
        self.payload = dict(self.query.items())
        assert "session" in context, "No database session passed to query parser"
        self.session = context["session"]

        self._validate_having_fields()
        self._apply_or()
        self._apply_and()
        self._apply_resource_fields()

        return ParsingResult(result=self.payload, errors=self.errors)

    def _validate_having_fields(self) -> None:
        if self.query.get(HAVING_KEY, None):
            assert self.resource._aggregate, f"Aggregation is not allowed on this resource {self.resource}"
            for key, field in self.resource.having._fields.items():
                assert isinstance(
                    field, BaseScalarField
                ), f"Having Resources can only contain BaseScalarfield but got: {field} on key: {key}"

    def _apply_or(self) -> None:
        self._apply_listing("_or")

    def _apply_and(self) -> None:
        self._apply_listing("_and")

    def _apply_listing(self, key: str) -> None:
        if key not in self.query:
            return

        payload = self.query[key]
        if not self.expect_list(key, payload):
            return

        parsing_results = (self._parse_own(item) for item in payload)
        result_list: List[Mapping[str, Any]] = []
        for index, parsing_result in enumerate(parsing_results):
            result_list.append(parsing_result.result)
            if len(parsing_result.errors) > 0:
                errors = self.errors.setdefault(key, {})
                if isinstance(errors, dict):
                    errors[index] = parsing_result.errors

        self.payload[key] = result_list

    def _parse_own(self, item: Dict[str, Any]) -> ParsingResult:
        sub_parser = self.__class__(self.resource, item)
        return sub_parser.parse(self.context)

    def _apply_resource_fields(self) -> None:
        for key in self.resource._fields.keys():
            self._apply_resource_field(key)

    def _apply_resource_field(self, key: str) -> None:
        field = self.resource._fields[key]
        if not isinstance(field, fields.ConcreteResolver):
            # Ignore all fields which are not resolvers for processing
            # of partial data.
            return
        assert isinstance(field, fields.ConcreteResolver)

        try:
            query_key = self._get_existing_key(key)
        except KeyError:
            return

        query_config = self.query[query_key]
        field_parser = FieldParser(key, query_config, field)
        result = field_parser.parse(self.context)
        sub_errors = result.errors

        del self.payload[query_key]
        if len(sub_errors):
            self.errors[query_key] = cast(NestedErrorItem, sub_errors)
            # As there might be a huge payload from the database, aborting here
            # to avoid an easy DDOS path
            return

        primary_key_column = field.nested_resource.primary_key_column
        query = self.session.query(primary_key_column)
        query = self._apply_domain_context(query, field)
        query = result.apply_to_query(query)
        primary_keys = [item[0] for item in query.all()]
        request = ResolverRequest(
            attribute_key=query_key, sub_resolver=field, primary_keys=primary_keys, parent=self.payload
        )
        # Happens inline. This is just set ot the payload again to reduce confusion.
        self.payload = field.resolver_function(request)

    def _apply_domain_context(self, query: Query, field: fields.ConcreteResolver) -> Query:
        domain_column = getattr(field.nested_resource.model, "domain_uuid", None)
        domain_context = self.context.get("domain_uuid")
        if domain_context is not None and domain_column is not None:
            if isinstance(domain_context, (tuple, list)):
                filter_entry = domain_column.in_(domain_context)
            else:
                filter_entry = domain_column == domain_context
            return query.filter(filter_entry)
        return query
