from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Dict, Generic, Iterable, Union

from sqlalchemy import inspect
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.sql.elements import ColumnClause
from sqlalchemy.sql.schema import Column

from xcnt.sqlalchemy.search.parser.resource_base import ResourceBaseParser, ResourceBaseQueryType
from xcnt.sqlalchemy.search.parser.resource_reference import ResourceReferenceParserMixin

if TYPE_CHECKING:
    from xcnt.sqlalchemy.search.fields.base import BaseField
    from xcnt.sqlalchemy.search.query import HavingBaseClass, Resource


class ResourceParser(
    ResourceBaseParser[ResourceBaseQueryType],  # type: ignore
    ResourceReferenceParserMixin,
    Generic[ResourceBaseQueryType],  # type: ignore
    ABC,
):
    def __init__(self, resource: Resource):
        ResourceReferenceParserMixin.__init__(self, resource)
        ResourceBaseParser.__init__(self)

    @property
    def resource_fields(self) -> Dict[str, BaseField]:
        return self.resource._fields

    @property
    def model(self) -> Union[DeclarativeMeta, AliasedClass]:
        return self.resource.model

    @property
    def having(self) -> HavingBaseClass:
        return self.resource.having

    @property
    def having_fields(self) -> Dict[str, BaseField]:
        return self.having._fields

    @property
    def _primary_key_column(self) -> ColumnClause:
        return self.resource.primary_key_column

    @property
    def _primary_key_column_name(self) -> str:
        return self.resource.primary_key_column_name

    @property
    def _ignored_fields(self) -> Iterable[str]:
        return self.resource.ignored_fields

    @property
    def root_model(self) -> DeclarativeMeta:
        model = self.model
        if isinstance(model, AliasedClass):
            return inspect(model).mapper.class_
        else:
            return model

    @property
    def _root_model_primary_key_column(self) -> Column:
        return getattr(self.root_model, self._primary_key_column_name)

    def _ignore_key(self, key: str) -> bool:
        return key in self._ignored_fields or super()._ignore_key(key)
