from abc import ABC, abstractproperty
from typing import Any, Dict, Generic, Iterable, Iterator, Mapping, Tuple, TypeVar, Union

from xcnt.sqlalchemy.search.fields import BaseField, ComputedField, ConcreteBase
from xcnt.sqlalchemy.search.parser.evaluation_result import (
    EvaluationErrorCode,
    EvaluationResult,
    EvaluationResultParser,
    key_derivations,
)
from xcnt.sqlalchemy.search.parser.field import FieldParser
from xcnt.sqlalchemy.search.parser.flow_exception import ReturnAll

ResourceBaseQueryType = TypeVar("ResourceBaseQueryType", bound=Mapping[str, Any])


class ResourceBaseParser(EvaluationResultParser, Generic[ResourceBaseQueryType], ABC):
    query: ResourceBaseQueryType

    def _parse_fields_or_raise_nothing(self) -> None:
        self.sub_results.extend(self._parse_fields())

    def _parse_fields(self) -> Iterable[EvaluationResult]:
        for key, query_config, resource_field in self.query_fields:
            if not isinstance(resource_field, ConcreteBase):
                raise AttributeError(
                    f"Field with key {key} does not point to a ConcreteBase field. "
                    f"Resources don't support other. Got: {resource_field}"
                )
            field_parser = FieldParser(key, query_config, resource_field)
            try:
                result = field_parser.parse(self.parse_context)
                yield result
            except ValueError:
                if isinstance(resource_field, ComputedField):
                    nested_field_name = resource_field.field.__class__.__name__.lower()
                    error = {key: EvaluationErrorCode.INVALID_VALUE_FOR_FILTER}
                    self.errors[nested_field_name] = error  # type:ignore
                else:
                    self.errors[key] = EvaluationErrorCode.INVALID_VALUE_FOR_FILTER
            except ReturnAll:
                pass

    @property
    def query_fields(self) -> Iterator[Tuple[str, Union[Dict[str, Any], Any], BaseField]]:
        if not issubclass(self.query.__class__, Mapping):
            self.errors["query_fields"] = EvaluationErrorCode.NOT_A_DICT
        else:
            for key, value in self.query.items():
                if self._ignore_key(key):
                    continue

                try:
                    resource_key = self._get_resource_key(key)
                except KeyError:
                    resource_key = key

                if resource_key not in self.resource_fields:
                    self.errors[key] = EvaluationErrorCode.UNKNOWN_FIELD
                    continue

                resource_field = self.resource_fields[resource_key]
                if not isinstance(value, dict) and not resource_field.accepts_non_dict:
                    self.errors[key] = EvaluationErrorCode.NOT_A_DICT
                else:
                    resource_field = self.resource_fields[resource_key]
                    yield key, value, resource_field

    def _get_resource_key(self, key: str) -> str:
        for check_key in key_derivations(key):
            if check_key in self.resource_fields:
                return check_key
        raise KeyError(f'Resource key "{key}" not found')

    @abstractproperty
    def resource_fields(self) -> Dict[str, BaseField]:
        pass

    @property
    def parse_context(self) -> Dict[str, Any]:
        return getattr(self, "_parse_context", {})

    @parse_context.setter
    def parse_context(self, context: Dict[str, Any]) -> None:
        self._parse_context = context

    def _ignore_key(self, key: str) -> bool:
        return key.startswith("_")
