from __future__ import annotations

from abc import ABC
from typing import Generic, TypeVar

from sqlalchemy import inspect
from sqlalchemy.orm.descriptor_props import ConcreteInheritedProperty
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.sql.elements import ColumnElement

from xcnt.sqlalchemy.search.fields.function_scalar_base import FunctionScalarBase
from xcnt.sqlalchemy.search.parser.field_plugin import FieldPlugin, QueryType
from xcnt.sqlalchemy.search.parser.filters import ModelFieldDoesNotExist
from xcnt.sqlalchemy.search.parser.flow_exception import ReturnNothing

BaseScalarFunctionField = TypeVar("BaseScalarFunctionField", bound=FunctionScalarBase)


class ScalarFunctionFieldParser(
    FieldPlugin[BaseScalarFunctionField, QueryType], Generic[BaseScalarFunctionField, QueryType], ABC  # type: ignore
):
    @property
    def model_field(self) -> ColumnElement:
        field = self.resource_field

        if field.table_column_name is not None:
            return self.table_field

        field_name = field.name
        assert field_name is not None

        try:
            model_field = self.get_column(field_name)
            original_property = getattr(model_field, "original_property", None)
            if original_property is not None and isinstance(original_property, ConcreteInheritedProperty):
                column = getattr(self.model.__table__.c, field_name, None)
                if column is not None:
                    return column
                raise ReturnNothing()
            if model_field is None:
                raise AttributeError("Did not find element")
            return model_field
        except AttributeError as error:
            raise ModelFieldDoesNotExist(f"No field {field.name} on {self.model}", field.name, self.model) from error

    @property
    def table_field(self) -> ColumnElement:
        field = self.resource_field
        table_column_name = field.table_column_name
        assert table_column_name is not None

        # This is required to allow support for proper
        # aliasing in sub queries. It first checks if the column
        # name can be accessed via the model and only if not
        # falls back to the table.
        if isinstance(self.model, AliasedClass):
            table = inspect(self.model).selectable
        else:
            table = self.model.__table__

        try:
            return getattr(table.c, table_column_name)
        except AttributeError:
            raise ModelFieldDoesNotExist(
                f"No field {table_column_name} on table {self.model.__table__}", field.name, self.model
            )
