from __future__ import annotations

from typing import Any, Callable, Dict, Iterable, Optional, Type, cast

from sqlalchemy import literal
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.elements import ClauseElement, ColumnClause

from xcnt.sqlalchemy.search.meta_resolver import find_meta_key
from xcnt.sqlalchemy.search.parser import (
    BaseQueryParser,
    EvaluationResult,
    NestedQueryParser,
    QueryParser,
    ReturnAll,
    ReturnNothing,
)
from xcnt.sqlalchemy.search.query.resource_base import ResourceBase, ResourceBaseMeta
from xcnt.sqlalchemy.search.query.resource_having import DefaultHavingClass, HavingBaseClass


class ResourceMeta(ResourceBaseMeta):
    _meta_key: Callable[[Any, str], str]

    @property
    def model(self) -> DeclarativeMeta:
        return self._meta_key("model")

    @property
    def primary_key_column_name(self) -> str:
        try:
            primary_key_column_name = self._meta_key("primary_key_column_name")
        except KeyError:
            primary_key_column_name = inspect(self.model).primary_key[0].name
        return primary_key_column_name

    @property
    def primary_key_column(self) -> ColumnClause:
        return getattr(self.model, self.primary_key_column_name)

    @property
    def having(self) -> Type[HavingBaseClass]:
        return self._meta_key("having")  # type: ignore

    @property
    def _aggregate(self) -> bool:
        return self._meta_key("aggregate")  # type: ignore


def get_column_of(model: DeclarativeMeta, column_name: str) -> Optional[ClauseElement]:
    return getattr(model, column_name, None)


ClauseElementGetter = Callable[[DeclarativeMeta, str], Optional[ClauseElement]]


class Resource(ResourceBase, metaclass=ResourceMeta):
    class Meta:
        model: DeclarativeMeta
        abstract = True
        aggregate = False
        clause_element_getter: ClauseElementGetter = get_column_of
        ignored_fields: Iterable[str] = []
        having: Type[HavingBaseClass] = DefaultHavingClass

    def __init__(
        self, model: Optional[DeclarativeMeta] = None, clause_element_getter: Optional[ClauseElementGetter] = None
    ):
        self.model = model or self._meta_key("model")
        super().__init__()
        having_class = self._meta_key("having")
        self.having = having_class(self)
        aggregate = self._meta_key("aggregate")
        self.clause_element_getter = clause_element_getter or cast(
            ClauseElementGetter, self._meta_key("clause_element_getter")
        )
        self._aggregate = aggregate

    def copy_for_model(self, model: DeclarativeMeta) -> Resource:
        return self.__class__(model=model, clause_element_getter=self.clause_element_getter)

    @classmethod
    def _meta_key(self, key: str) -> Any:
        return find_meta_key(self, key)

    def get_column(self, column_name: str) -> Optional[ClauseElement]:
        return self.clause_element_getter(self.model, column_name)

    @property
    def ignored_fields(self) -> Iterable[str]:
        return self._meta_key("ignored_fields")

    @property
    def primary_key_column_name(self) -> str:
        return cast(str, self.__class__.primary_key_column_name)

    @property
    def primary_key_column(self) -> ClauseElement:
        primary_key_column = self.get_column(self.primary_key_column_name)
        assert (
            primary_key_column is not None
        ), f"Expected a primary key column of name {self.primary_key_column_name} on {self.model}. Got nothing."
        return primary_key_column

    def apply(self, query: Query, query_filter: Dict[str, Any], context: Optional[Dict[str, Any]] = None) -> Query:
        generated = self.generate(query_filter, context)
        return generated.apply_to_query(query)

    def generate(self, query_filter: Dict[str, Any], context: Optional[Dict[str, Any]] = None) -> EvaluationResult:
        query_parser = QueryParser(query_filter, self)
        return self._generate_with(query_parser, context)

    def generate_nested(
        self, query_filter: Dict[str, Any], context: Optional[Dict[str, Any]] = None
    ) -> EvaluationResult:
        query_parser = NestedQueryParser(query_filter, self)
        return self._generate_with(query_parser, context)

    def _generate_with(
        self, query_parser: BaseQueryParser, context: Optional[Dict[str, Any]] = None
    ) -> EvaluationResult:
        try:
            return self._generate(query_parser, context)
        except ReturnAll:
            return EvaluationResult.empty()
        except ReturnNothing:
            return EvaluationResult(expressions=[literal(False)])

    def _generate(self, query_parser: BaseQueryParser, context: Optional[Dict[str, Any]] = None) -> EvaluationResult:
        context = context or {}
        return query_parser.parse(context)
