from __future__ import annotations

import copy
import weakref
from typing import TYPE_CHECKING, Any, Dict, MutableMapping, Optional, Tuple, Type, Union, cast

from xcnt.sqlalchemy.search import registry
from xcnt.sqlalchemy.search.fields.base import BaseField
from xcnt.sqlalchemy.search.fields.concrete_base import ConcreteBase
from xcnt.sqlalchemy.search.meta_resolver import direct_meta_key

if TYPE_CHECKING:
    from xcnt.sqlalchemy.search.query.resource import Resource


def is_abstract(check_class: Type) -> bool:
    try:
        return bool(direct_meta_key(check_class, "abstract"))
    except KeyError:
        return False


def _generate_fields_of(item: Any, resource: Optional[Resource] = None) -> Dict[str, BaseField]:
    fields: Dict[str, BaseField] = {}
    for key in dir(item):
        if key.startswith("_"):
            continue

        try:
            attribute = getattr(item, key)
        except Exception:
            # It seems like the attribute is not accessible. Most probably not a field. Continuing
            continue

        if isinstance(attribute, BaseField):
            attribute = copy.copy(attribute)

        if isinstance(attribute, ConcreteBase):
            if resource is None:
                attribute.resource = item
            else:
                attribute.resource = resource

        if isinstance(attribute, BaseField):
            if attribute.name is None:
                attribute.name = key

            if attribute.query_field_name is not None:
                key = attribute.query_field_name

            fields[key] = attribute

    return fields


# This is required to ensure that the fields are not being shared
# between the class and the instance. This results
# if you create an apispec and do a query afterwards in an error.
# Not really happy with this solution but to make everything
# workable and allowing separate fields on classes and instances
# this seems to be the least terrible option (please prove me wrong)
fields_of_items: MutableMapping[Union[FieldGeneratorBaseMixin], Dict[str, BaseField]] = weakref.WeakKeyDictionary()


class FieldGeneratorBaseMixin:
    @property
    def _fields(self) -> Dict[str, BaseField]:
        if self not in fields_of_items:
            fields_of_items[self] = _generate_fields_of(self, resource=self._resource_for_fields)
        return fields_of_items[self]

    @property
    def _resource_for_fields(self) -> Optional[Resource]:
        return None


class ResourceBaseMeta(type, FieldGeneratorBaseMixin):
    def __init__(cls, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any]) -> None:
        super().__init__(name, bases, attrs)

        if name and not is_abstract(cls):
            registry.register(name, cast(Type[ResourceBase], cls))

        FieldGeneratorBaseMixin.__init__(cls)


class ResourceBase(FieldGeneratorBaseMixin, metaclass=ResourceBaseMeta):
    class Meta:
        abstract = True

    @property
    def _name(self) -> str:
        return self.__class__.__name__
