from __future__ import annotations

import itertools
from typing import TYPE_CHECKING, Dict, Iterable, List, Type

from xcnt.sqlalchemy.search.exceptions import RegistryError

if TYPE_CHECKING:
    from xcnt.sqlalchemy.search.query.resource_base import ResourceBase

    ResourceType = Type[ResourceBase]


class Registry:
    _registry: Dict[str, List[ResourceType]]

    def __init__(self) -> None:
        self._registry = {}

    def register(self, classname: str, resource: ResourceType) -> None:
        registry = self._registry
        module = resource.__module__
        fullpath = ".".join([module, classname])
        if classname in registry and not any(check.__module__ == module for check in registry[classname]):
            registry[classname].append(resource)
        elif classname not in self._registry:
            registry[classname] = [resource]

        registry[fullpath] = [resource]

    def get_class(self, classname: str) -> ResourceType:
        classes = self.get_classes(classname)
        if len(classes) == 1:
            return classes[0]
        else:
            raise RegistryError(
                f"Multiple classes with name {classname} were found. Please use the full, module-qualified path."
            )

    def get_classes(self, classname: str) -> List[ResourceType]:
        try:
            return self._registry[classname]
        except KeyError as error:
            raise RegistryError(
                f"Class with name {classname} was not found. You may need to import the class."
            ) from error

    def get_all(self) -> Iterable[ResourceType]:
        return itertools.chain.from_iterable(self._registry.values())


_registry = Registry()
register = _registry.register
get_class = _registry.get_class
get_classes = _registry.get_classes
get_all = _registry.get_all
