import pytest
from apispec import APISpec
from apispec.utils import build_reference

from xcnt.sqlalchemy.search.query.resource_having import DefaultHavingClass
from xcnt.sqlalchemy.search.test import query
from xcnt.sqlalchemy.search.test.query.datapoint_float import HavingDatapoint


@pytest.fixture
def parameterized_api_spec(api_spec: APISpec):
    parameter = {"content": {"application/json": {"schema": query.Component.__name__}}}
    api_spec.components.parameter("where", "query", parameter)
    return api_spec


def test_plugin_usage(parameterized_api_spec: APISpec, get_parameters) -> None:
    api_spec = parameterized_api_spec
    params = get_parameters()
    schema_reference = params["where"]["content"]["application/json"]["schema"]
    assert schema_reference == build_reference("schema", api_spec.openapi_version.major, "ComponentQuery")

    components_dict = api_spec.components.to_dict()
    assert "ComponentQuery" in components_dict["schemas"]
    schema = components_dict["schemas"]["ComponentQuery"]
    assert schema["properties"]["code"]["properties"]["_in"]["type"] == "array"

    assert schema["properties"]["component_model"] == build_reference(
        "schema", api_spec.openapi_version.major, "NestedComponentModelQuery"
    )


@pytest.mark.parametrize("key", ["_and", "_or"])
def test_and_or_spec(key: str, parameterized_api_spec: APISpec, get_parameters) -> None:
    get_parameters()
    api_spec = parameterized_api_spec

    components_dict = api_spec.components.to_dict()
    assert "ComponentQuery" in components_dict["schemas"]
    schema = components_dict["schemas"]["ComponentQuery"]
    assert schema["properties"][key]["items"] == build_reference(
        "schema", api_spec.openapi_version.major, "ComponentQuery"
    )


def test_having_spec(api_spec: APISpec) -> None:
    parameter = {"content": {"application/json": {"schema": query.DatapointFloat.__name__}}}
    api_spec = api_spec.components.parameter("where", "query", parameter)

    components_dict = api_spec.to_dict()
    assert "DatapointFloatQuery" in components_dict["schemas"]
    schema = components_dict["schemas"]["DatapointFloatQuery"]
    assert schema["properties"]["_having"] == build_reference(
        "schema", api_spec.openapi_version.major, HavingDatapoint.__name__
    )


def test_not_having_root_spec(api_spec: APISpec) -> None:
    parameter = {"content": {"application/json": {"schema": query.System.__name__}}}
    api_spec = api_spec.components.parameter("where", "query", parameter)

    components_dict = api_spec.to_dict()
    assert "SystemQuery" in components_dict["schemas"]
    schema = components_dict["schemas"]["SystemQuery"]
    assert "_having" not in schema["properties"]


def test_having_nested_spec(api_spec: APISpec) -> None:
    parameter = {"content": {"application/json": {"schema": query.System.__name__}}}
    api_spec = api_spec.components.parameter("where", "query", parameter)

    components_dict = api_spec.to_dict()
    assert "NestedComponentQuery" in components_dict["schemas"]
    schema = components_dict["schemas"]["NestedComponentQuery"]
    assert schema["properties"]["_having"] == build_reference(
        "schema", api_spec.openapi_version.major, DefaultHavingClass.__name__
    )


def test_having_nested_reference(api_spec: APISpec):
    parameter = {"content": {"application/json": {"schema": query.System.__name__}}}
    api_spec = api_spec.components.parameter("where", "query", parameter)

    components_dict = api_spec.to_dict()
    assert "SystemQuery" in components_dict["schemas"]
    schema = components_dict["schemas"]["SystemQuery"]
    assert schema["properties"]["components"] == build_reference(
        "schema", api_spec.openapi_version.major, "NestedComponentQuery"
    )


def test_having_groups(api_spec: APISpec):
    parameter = {"content": {"application/json": {"schema": query.DatapointFloat.__name__}}}
    api_spec = api_spec.components.parameter("where", "query", parameter)

    datapoint_dict = api_spec.to_dict()
    assert "DatapointFloatQuery" in datapoint_dict["schemas"]
    schema = datapoint_dict["schemas"]["DatapointFloatQuery"]
    assert schema["properties"]["_having"] == build_reference(
        "schema", api_spec.openapi_version.major, HavingDatapoint.__name__
    )

    assert HavingDatapoint.__name__ in datapoint_dict["schemas"]
    having_schema = datapoint_dict["schemas"][HavingDatapoint.__name__]
    assert having_schema["properties"]["nested_having_group"] == build_reference(
        "schema", api_spec.openapi_version.major, "NestedHavingAggregateDatapoint"
    )

    assert "NestedHavingAggregateDatapoint" in datapoint_dict["schemas"]
    having_nested_schema = datapoint_dict["schemas"]["NestedHavingAggregateDatapoint"]
    assert having_nested_schema["properties"]["avg"] is not None


@pytest.mark.parametrize("key", ["_and", "_or", "_one_of", "_not_of"])
def test_having_no_and_or(api_spec: APISpec, key: str):
    parameter = {"content": {"application/json": {"schema": query.DatapointFloat.__name__}}}
    api_spec = api_spec.components.parameter("where", "query", parameter)

    datapoint_dict = api_spec.to_dict()
    assert "DatapointFloatQuery" in datapoint_dict["schemas"]
    schema = datapoint_dict["schemas"]["DatapointFloatQuery"]
    assert schema["properties"][key] is not None

    assert HavingDatapoint.__name__ in datapoint_dict["schemas"]
    having_schema = datapoint_dict["schemas"][HavingDatapoint.__name__]
    assert having_schema["properties"].get(key, None) is None

    assert "NestedHavingAggregateDatapoint" in datapoint_dict["schemas"]
    having_nested_schema = datapoint_dict["schemas"]["NestedHavingAggregateDatapoint"]
    assert having_nested_schema["properties"].get(key, None) is None


def test_field_group(api_spec: APISpec):
    parameter = {"content": {"application/json": {"schema": query.Component.__name__}}}
    api_spec = api_spec.components.parameter("where", "query", parameter)
    result = api_spec.to_dict()
    assert "aggregate" in result["schemas"]["ComponentQuery"]["properties"]
    assert result["schemas"]["ComponentQuery"]["properties"]["aggregate"] == build_reference(
        "schema", api_spec.openapi_version.major, "DatapointAggregateGroup"
    )
    assert "DatapointAggregateGroup" in result["schemas"]
    aggregate_group_schema = result["schemas"]["DatapointAggregateGroup"]
    assert "float" in aggregate_group_schema["properties"]


def test_function_field(api_spec: APISpec):
    parameter = {"content": {"application/json": {"schema": query.Datapoint.__name__}}}
    api_spec = api_spec.components.parameter("where", "query", parameter)
    result = api_spec.to_dict()
    datapoint_query_attributes = result["schemas"]["DatapointQuery"]["properties"]
    assert "start" in datapoint_query_attributes
    assert datapoint_query_attributes["start"]["type"] == "string"
