from enum import Enum
from typing import Any, Dict

from sqlalchemy import Column, DateTime, Float, ForeignKey, LargeBinary, String, sql
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.declarative import AbstractConcreteBase, declared_attr
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, relationship
from sqlalchemy.sql.expression import ClauseElement

from xcnt.sqlalchemy.search.test.model.base import Base
from xcnt.sqlalchemy.search.test.model.component import Component


class DataTypeEnum(Enum):
    FLOAT = "float"
    STRING = "string"
    BLOB = "blob"


def type_column(data_type: DataTypeEnum) -> hybrid_property:
    @hybrid_property
    def type_property(self) -> DataTypeEnum:
        return data_type

    @type_property.expression  # type: ignore
    def type_property(self) -> ClauseElement:
        return sql.expression.label("type", sql.literal(data_type.value))

    return type_property


class HasComponentMixin:
    __tablename__: str

    @declared_attr
    def component_uuid(self) -> Column:
        return Column(
            UUID(as_uuid=True),
            ForeignKey(
                "components.uuid",
                name=f"{self.__tablename__}_component_uuid",
            ),
            nullable=True,
            index=True,
        )

    @declared_attr
    def component(self) -> relationship:
        return relationship(
            Component,
            uselist=False,
            backref=backref(self.__tablename__, uselist=True),
            foreign_keys=[self.component_uuid],
        )


class DatapointBase(AbstractConcreteBase, Base):
    time = Column(DateTime, nullable=False)


class DatapointFloat(DatapointBase, HasComponentMixin):
    __tablename__ = "float_datapoints"

    value = Column("value_float", Float)

    type = type_column(DataTypeEnum.FLOAT)

    @declared_attr
    def __mapper_args__(cls) -> Dict[str, Any]:
        data = {"polymorphic_identity": "float", "concrete": True}
        return data


class DatapointString(DatapointBase, HasComponentMixin):
    __tablename__ = "string_datapoints"

    value = Column("value_string", String)

    type = type_column(DataTypeEnum.STRING)

    @declared_attr
    def __mapper_args__(cls) -> Dict[str, Any]:
        data = {"polymorphic_identity": "string", "concrete": True}
        return data


class DatapointBlob(DatapointBase, HasComponentMixin):
    __tablename__ = "blob_datapoints"

    value = Column("value_blob", LargeBinary)

    type = type_column(DataTypeEnum.BLOB)

    @declared_attr
    def __mapper_args__(cls) -> Dict[str, Any]:
        data = {"polymorphic_identity": "blob", "concrete": True}
        return data
