from typing import Any, Dict, List, Mapping, Optional, Type, Union
from uuid import UUID, uuid4

import pytest

from xcnt.sqlalchemy.search import EvaluationErrorCode
from xcnt.sqlalchemy.search.parser.partial import ParsingResult
from xcnt.sqlalchemy.search.test import query as query_configurations


@pytest.fixture
def component(factories):
    return factories.Component()


@pytest.fixture
def configuration_class():
    return query_configurations.PartialComponent


@pytest.fixture
def result_for(session, configuration_class: Type[query_configurations.PartialComponent]):
    def _result_for(
        result_filter: Dict[str, Any], domain_uuid: Optional[UUID] = None, domain_uuid_as_list: bool = False
    ) -> Mapping[str, Any]:
        configuration = configuration_class()
        context = {"session": session}
        domain_uuid_to_apply: Optional[Union[UUID, List[UUID]]]
        if domain_uuid is not None:
            if domain_uuid_as_list:
                domain_uuid_to_apply = [domain_uuid]
            else:
                domain_uuid_to_apply = domain_uuid
            context["domain_uuid"] = domain_uuid_to_apply
        return configuration.generate_partial_resolved_dict(result_filter, context=context)

    return _result_for


@pytest.fixture
def partial_result_for(session, configuration_class: Type[query_configurations.PartialComponent]):
    def _partial_result_for(result_filter: Dict[str, Any]) -> ParsingResult:
        configuration = configuration_class()
        return configuration.generate(result_filter, context={"session": session})

    return _partial_result_for


def test_nested_eq(result_for, component, factories):
    factories.Component()
    result_query = result_for({"system": {"alias": {"_eq": component.system.alias}}})
    assert result_query["system"]["_oneOf"] == [component.system.uuid]


def test_nested_morphed_eq(result_for, component, factories):
    factories.Component()
    result_query = result_for({"componentModel": {"alias": {"_eq": component.component_model.alias}}})
    assert result_query["componentModel"]["_oneOf"] == [component.component_model.uuid]


def test_having_filter_disabled_aggregation(result_for, component, factories):
    with pytest.raises(AssertionError):
        result_for(
            {"_having": {"count": {"_eq": 2}}, "componentModel": {"alias": {"_eq": component.component_model.alias}}}
        )


def test_having_filter_invalid_field(session, component, factories):
    configuration = query_configurations.InvalidAggregationPartialComponent()
    context = {"session": session}
    result_filter = {
        "_having": {"count": {"_eq": 2}},
        "componentModel": {"alias": {"_eq": component.component_model.alias}},
    }

    with pytest.raises(AssertionError):
        configuration.generate_partial_resolved_dict(query=result_filter, context=context)


@pytest.fixture
def nested_duplication_query_dict(component):
    return {
        "_and": [
            {
                "system": {"alias": {"_eq": component.system.alias}},
                "_and": [{"system": {"name": {"_eq": component.system.name}}}],
            },
            {"system": {"description": {"_eq": component.system.description}}},
        ]
    }


def test_nested_deduplication(result_for, nested_duplication_query_dict, component, factories, session, mocker):
    system2 = factories.System(alias=component.system.alias)
    result_query = result_for(nested_duplication_query_dict)
    assert sorted(result_query["_and"][0]["system"]["_oneOf"]) == sorted([component.system.uuid, system2.uuid])
    assert sorted(result_query["_and"][0]["_and"][0]["system"]["_oneOf"]) == sorted([component.system.uuid])
    assert sorted(result_query["_and"][1]["system"]["_oneOf"]) == sorted([component.system.uuid])


def test_nested_deduplication_no_error(partial_result_for, nested_duplication_query_dict, factories):
    result = partial_result_for(nested_duplication_query_dict)
    assert result.errors == {}


def test_nested_duplication_error(partial_result_for, component):
    result = partial_result_for(
        {
            "_and": [
                {
                    "system": {"alias": {"_eq": component.system.alias}},
                    "_and": {"system": {"name": {"_eq": component.system.name}}},
                },
                {"system": {"description": {"_eq": component.system.description}}},
            ]
        }
    )
    assert result.errors["_and"][0]["_and"] == EvaluationErrorCode.NOT_A_LIST


@pytest.mark.parametrize("configuration_class", [query_configurations.PartialComponentWithSpecialKey])
def test_nested_with_custom_resolve(result_for, component, factories):
    factories.Component()
    result_query = result_for({"system": {"alias": {"_eq": component.system.alias}}})
    assert result_query["system_uuid"]["_in"] == [component.system.uuid]


@pytest.mark.parametrize("domain_uuid_as_list", [False, True])
def test_apply_domain_context_query(result_for, factories, domain_uuid_as_list):
    domain_uuid = uuid4()
    factories.System.create_batch(8)
    system = factories.System(domain__uuid=domain_uuid)
    result_query = result_for(
        {"system": {"alias": {"_eq": system.alias}}}, domain_uuid, domain_uuid_as_list=domain_uuid_as_list
    )
    assert result_query["system"]["_oneOf"] == [system.uuid]


def test_apply_domain_context_invalid_domain(result_for, factories):
    domain_uuid = uuid4()
    system = factories.System()
    result_query = result_for({"system": {"alias": {"_eq": system.alias}}}, domain_uuid)
    assert result_query["system"]["_oneOf"] == []
