from datetime import datetime, timedelta
from typing import Any, Dict
from uuid import uuid4

import pytest
from sqlalchemy.orm.query import Query, aliased

from xcnt.sqlalchemy.search import EvaluationErrorCode
from xcnt.sqlalchemy.search.test import model
from xcnt.sqlalchemy.search.test import query as query_configurations


@pytest.fixture
def system(factories):
    return factories.System()


@pytest.fixture
def systems(factories):
    return factories.System.create_batch(20)


@pytest.fixture
def query_object():
    return model.System


@pytest.fixture
def result_for(session, query_object):
    def _result_for(result_filter: Dict[str, Any]) -> Query:
        configuration = query_configurations.System(query_object)
        return configuration.generate(result_filter)

    return _result_for


@pytest.fixture
def query(session, query_object):
    def _query(query_filter: Dict[str, Any]) -> Query:
        query = session.query(query_object)
        configuration = query_configurations.System(query_object)
        return configuration.apply(query, query_filter)

    return _query


@pytest.mark.parametrize("query_object", [model.System, aliased(model.System)])
def test_eq(query, system, factories):
    factories.System()
    db_query = query({"alias": {"_eq": system.alias}})
    result = db_query.all()

    assert len(result) == 1
    result[0].uuid == system


def test_aliased_eq(query, system, factories):
    factories.System()
    db_query = query({"alias": {"_eq": system.alias}})
    result = db_query.all()

    assert len(result) == 1
    result[0].uuid == system


def test_neq(query, system, factories):
    other_system = factories.System()
    result = query({"alias": {"_neq": other_system.alias}}).all()

    assert len(result) == 1
    result[0].uuid == system


def check_for_presence(result, expected_items, not_expected_items):
    assert len(result) == len(expected_items)

    result_uuids = [item.uuid for item in result]
    for expected_item in expected_items:
        assert expected_item.uuid in result_uuids

    for not_expected_item in not_expected_items:
        assert not_expected_item.uuid not in result_uuids


def test_in(query, system, factories, systems):
    query_object = query({"alias": {"_in": [system.alias for system in systems[:10]]}})
    result = query_object.all()

    check_for_presence(result, systems[:10], systems[10:])


def test_not_in(query, factories, systems):
    result = query({"alias": {"_nin": [system.alias for system in systems[:10]]}}).all()

    assert len(result) == 10

    check_for_presence(result, systems[10:], systems[:10])


@pytest.mark.parametrize("key", ["_in", "_nin"])
def test_list_none(query, key, factories, systems):
    factories.System.create(alias=None)
    query_object = query({"alias": {key: None}})
    result = query_object.all()

    assert len(result) == 21


@pytest.mark.parametrize("key", ["_in", "_nin"])
def test_list_empty(query, key, factories, systems):
    query_object = query({"alias": {key: []}})
    result = query_object.all()

    assert len(result) == 0


def test_lt(query, factories):
    system = factories.System(created_at=datetime(2020, 1, 20))
    system2 = factories.System(created_at=system.created_at + timedelta(hours=2))

    result = query({"created_at": {"_lt": system2.created_at}}).all()
    assert len(result) == 1
    retrieved = result[0]
    assert system.uuid == retrieved.uuid


def test_lt_with_none_value(query, factories):
    system = factories.System(created_at=datetime(2020, 1, 20))
    system2 = factories.System(created_at=system.created_at + timedelta(hours=2))
    factories.System(created_at=system2.created_at + timedelta(hours=1))

    result = query({"created_at": {"_lt": None}}).all()
    assert len(result) == 3


def test_lte(query, factories):
    system = factories.System(created_at=datetime(2020, 1, 20))
    system2 = factories.System(created_at=system.created_at + timedelta(hours=2))
    factories.System(created_at=system2.created_at + timedelta(hours=1))

    result = query({"created_at": {"_lte": system2.created_at}}).all()
    assert len(result) == 2


def test_lte_with_none_value(query, factories):
    system = factories.System(created_at=datetime(2020, 1, 20))
    system2 = factories.System(created_at=system.created_at + timedelta(hours=2))
    factories.System(created_at=system2.created_at + timedelta(hours=1))

    result = query({"created_at": {"_lte": None}}).all()
    assert len(result) == 3


def test_gt(query, factories):
    system = factories.System(created_at=datetime(2020, 1, 20))
    system2 = factories.System(created_at=system.created_at + timedelta(hours=2))

    result = query({"created_at": {"_gt": system.created_at.isoformat()}}).all()
    assert len(result) == 1
    retrieved = result[0]
    assert system2.uuid == retrieved.uuid


def test_gt_with_none_value(query, factories):
    system = factories.System(created_at=datetime(2020, 1, 20))
    factories.System(created_at=system.created_at + timedelta(hours=2))

    result = query({"created_at": {"_gt": None}}).all()
    assert len(result) == 2


def test_gte(query, factories):
    system = factories.System(created_at=datetime(2020, 1, 20))
    factories.System(created_at=system.created_at + timedelta(hours=2))
    factories.System(created_at=system.created_at - timedelta(hours=1))

    result = query({"created_at": {"_gte": system.created_at.isoformat()}}).all()
    assert len(result) == 2


def test_gte_with_none_value(query, factories):
    system = factories.System(created_at=datetime(2020, 1, 20))
    factories.System(created_at=system.created_at + timedelta(hours=2))

    result = query({"created_at": {"_gte": None}}).all()
    assert len(result) == 2


def test_in_invalid_data(result_for):
    result = result_for({"alias": {"_in": "something"}})
    assert result.errors["alias"]["_in"] == EvaluationErrorCode.INVALID_VALUE_FOR_FILTER


def test_unsupported_function(result_for):
    result = result_for({"alias": {"_lt": "something"}})
    assert result.errors["alias"]["_lt"] == EvaluationErrorCode.UNSUPPORTED_FILTER_FUNCTION


def test_unknown_function(result_for):
    result = result_for({"alias": {"_ltt2020": "something"}})
    assert result.errors["alias"]["_ltt2020"] == EvaluationErrorCode.UNKNOWN_FILTER_FUNCTION


@pytest.mark.parametrize("key", ["_one_of", "_oneOf"])
def test_basic(query, factories, key, systems):
    result = query({key: [system.uuid for system in systems[0:10]]}).all()
    check_for_presence(result, systems[:10], systems[10:])


@pytest.mark.parametrize("key", ["_not_of", "_notOf"])
def test_basic_negate(query, factories, key, systems):
    result = query({key: [system.uuid for system in systems[0:10]]}).all()
    check_for_presence(result, systems[10:], systems[:10])


@pytest.mark.parametrize("key", ["_not_of", "_notOf", "_oneOf", "_one_of"])
def test_basic_no_list_one_of_not_of(key, result_for):
    result = result_for({key: uuid4()})
    assert result.errors[key] == EvaluationErrorCode.NOT_A_LIST


@pytest.mark.parametrize("key", ["_one_of", "_oneOf"])
def test_basic_list_none_one_of(query, factories, key, systems):
    result = query({key: None}).all()
    assert len(result) == 20


@pytest.mark.parametrize("key", ["_not_of", "_notOf"])
def test_basic_list_none_not_of(query, factories, key, systems):
    result = query({key: None}).all()
    assert len(result) == 0


@pytest.mark.parametrize("key", ["_one_of", "_oneOf"])
def test_basic_list_empty_one_of(query, factories, key, systems):
    result = query({key: []}).all()
    assert len(result) == 0


@pytest.mark.parametrize("key", ["_not_of", "_notOf"])
def test_basic_list_empty_not_of(query, factories, key, systems):
    result = query({key: []}).all()
    assert len(result) == 20


def test_multiple(query, factories, systems):
    result = query({"alias": {"_eq": systems[0].alias}, "_not_of": [systems[0].uuid]}).all()
    assert len(result) == 0


def test_multiple_with_explicit_and(query, factories, systems):
    result = query({"_and": [{"alias": {"_eq": systems[0].alias}}, {"_not_of": [systems[0].uuid]}]}).all()
    assert len(result) == 0


def test_multiple_with_or(query, factories, systems):
    result = query(
        {"_or": [{"alias": {"_eq": systems[0].alias}}, {"_not_of": [systems[0].uuid, systems[-1].uuid]}]}
    ).all()
    check_for_presence(result, systems[:19], systems[-1:])


def test_like(query, factories, systems):
    system = factories.System.create(alias="test-1234")
    result = query({"alias": {"_like": "test%"}}).all()
    assert len(result) == 1
    assert system.alias == result[0].alias


def test_like_case_mismatch(query, factories, systems):
    factories.System.create(alias="test-1234")
    assert query({"alias": {"_like": "TeST%"}}).count() == 0


def test_like_null(query, systems):
    assert query({"alias": {"_like": None}}).count() == 20


def test_ilike(query, factories, systems):
    system = factories.System.create(alias="test-1234")
    result = query({"alias": {"_ilike": "test%"}}).all()
    assert len(result) == 1
    assert system.alias == result[0].alias


def test_ilike_case_insentitive(query, factories, systems):
    factories.System.create(alias="test-1234")
    assert query({"alias": {"_ilike": "TeST%"}}).count() == 1


def test_ilike_mismatch(query, factories, systems):
    factories.System.create(alias="test-1234")
    assert query({"alias": {"_ilike": "Te_ST%"}}).count() == 0


def test_ilike_null(query, systems):
    assert query({"alias": {"_ilike": None}}).count() == 20
