from datetime import datetime
from typing import Optional

import pytest
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy.sql.functions import func

from xcnt.sqlalchemy.search import fields, query
from xcnt.sqlalchemy.search.test import model
from xcnt.sqlalchemy.search.test import query as query_configurations


def test_datapoint_with_start_end(session, factories):
    point_1 = factories.DatapointFloat(time=datetime(2021, 4, 22, 0, 0, 0))
    point_2 = factories.DatapointFloat(time=datetime(2021, 4, 23, 12, 0, 0))
    point_3 = factories.DatapointFloat(time=datetime(2021, 4, 24, 23, 0, 0))
    configuration = query_configurations.Datapoint()
    query_filter = {"start": datetime(2021, 4, 23, 11, 0, 0), "end": datetime(2021, 4, 23, 12, 0, 1).isoformat()}
    query = session.query(model.DatapointBase)
    query = configuration.apply(query, query_filter)

    assert query.count() == 1
    datapoint = query.first()
    assert datapoint.uuid not in (point_1.uuid, point_3.uuid)
    assert datapoint.uuid == point_2.uuid


def test_function_group_by_system(session, factories):
    query = session.query(model.Component.system_uuid.label("system_uuid"), func.count().label("count"))
    system = factories.System()
    factories.Component(system=system)
    factories.Component(system=system)
    other_system = factories.System()
    factories.Component(system=other_system)

    configuration = query_configurations.ComponentGroup()
    query = configuration.apply(
        query,
        {
            "group_by_system": True,
        },
    )
    payload = query.all()
    assert len(payload) == 2
    lookup = {item.system_uuid: item.count for item in payload}
    assert lookup[system.uuid] == 2
    assert lookup[other_system.uuid] == 1


def test_grouping_function_no_grouping(session, factories):
    factories.Component.create_batch(5)
    query = session.query(func.count(model.Component.uuid).label("count"))
    configuration = query_configurations.ComponentGroup()
    query = configuration.apply(
        query,
        {
            "group_by_system": False,
        },
    )
    payload = query.all()
    assert len(payload) == 1
    assert payload[0].count == 5


def test_function_group_by_without_value_in_query(session, factories):
    system = factories.System()
    factories.Component(system=system)
    factories.Component(system=system)
    other_system = factories.System()
    factories.Component(system=other_system)

    query = session.query(func.count(model.Component.uuid).label("component_count"))
    configuration = query_configurations.ComponentGroup()
    query = configuration.apply(
        query,
        {
            "group_by_system": True,
        },
    )
    items = query.all()
    counts = [item.component_count for item in items]
    assert sorted(counts) == [1, 2]


def noop_boolean_function(model: DeclarativeMeta, column: ColumnElement, value: Optional[bool]) -> ColumnElement:
    return column


def test_function_field_with_unknown_target(session):
    class UnknownTargetQuery(query.Resource):
        class Meta:
            model = model.Component

        unknown_target_field = fields.BooleanFunction(  # type: ignore
            noop_boolean_function, attribute_name="uuid", target="Nonesense"
        )

    configuration = UnknownTargetQuery()
    with pytest.raises(NotImplementedError):
        configuration.generate({"unknown_target_field": False})
