import operator
from typing import Any, Dict

import pytest
from sqlalchemy.orm.query import Query

from xcnt.sqlalchemy.search.test import model
from xcnt.sqlalchemy.search.test import query as query_configurations


@pytest.fixture
def system(factories):
    return factories.System()


@pytest.fixture
def result_for(session):
    def _result_for(result_filter: Dict[str, Any]) -> Query:
        configuration = query_configurations.System()
        return configuration.generate(result_filter)

    return _result_for


@pytest.fixture
def query(session):
    def _query(query_filter: Dict[str, Any]) -> Query:
        query = session.query(model.System)
        configuration = query_configurations.System(model.System)
        return configuration.apply(query, query_filter)

    return _query


def test_having_count_subresource(query, system, factories):
    factories.System()

    factories.Component(system=system)
    factories.Component(system=system)
    db_query = query({"components": {"_having": {"count": {"_eq": 2}}}})
    result = db_query.all()

    components_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .right.element._having_criteria[0]
        .clauses[1]
    )
    assert components_having_stmt.left.name == "count"
    assert components_having_stmt.right.value == 2
    assert components_having_stmt.operator == operator.eq

    assert len(result) == 1
    result[0].uuid == system


def test_having_alias_count_subresource(query, system, factories):
    factories.System()

    factories.Component(system=system)
    factories.Component(system=system)
    db_query = query({"alias": {"_eq": system.alias}, "components": {"_having": {"count": {"_eq": 2}}}})

    system_alias_stmt = (
        db_query._where_criteria[0].right.element._raw_columns[0].element._where_criteria[0].clauses[0].clauses[1]
    )
    assert system_alias_stmt.left.key == "alias"
    assert system_alias_stmt.operator == operator.eq

    components_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .clauses[1]
        .right.element._having_criteria[0]
        .clauses[1]
    )
    assert components_having_stmt.left.name == "count"
    assert components_having_stmt.right.value == 2
    assert components_having_stmt.operator == operator.eq

    result = db_query.all()

    assert len(result) == 1
    result[0].uuid == system


def test_having_alias_double_count_subresource(session, system, factories):
    factories.Component(system__domain=factories.Domain())

    domain = factories.Domain()
    system.domain = domain
    factories.Component(system=system)
    factories.Component(system=system)

    system = factories.System(domain=domain)
    factories.Component(system=system)

    query = session.query(model.Domain)
    configuration = query_configurations.Domain(model.Domain)
    query_filter = {
        "systems": {
            "_having": {"count": {"_eq": 1}},
            "components": {
                "_having": {"count": {"_gte": 2}},
            },
        }
    }
    db_query = configuration.apply(query, query_filter)

    result = db_query.all()

    assert len(result) == 1
    result[0].uuid == domain.uuid


def test_having_with_joined_subquery(session, system, factories):
    factories.Component(system__domain=factories.Domain())

    domain = factories.Domain()
    system.domain = domain
    factories.Component(system=system)
    factories.Component(system=system)

    system = factories.System(domain=domain)
    component = factories.Component(system=system)

    query = session.query(model.Domain)
    configuration = query_configurations.Domain(model.Domain)
    query_filter = {
        "systems": {
            "_having": {"count": {"_eq": 1}},
            "components": {"componentModel": {"_oneOf": [component.component_model.uuid]}},
        }
    }
    db_query = configuration.apply(query, query_filter)

    result = db_query.all()

    assert len(result) == 1
    assert result[0].uuid == domain.uuid


def test_having_count_with_and_filter(query, system, factories):
    other_system = factories.System()
    factories.Component(system=other_system, code="test")

    factories.Component(system=system, code="test")
    factories.Component(system=system)
    factories.Component(system=system)
    db_query = query(
        {"_and": [{"components": {"_having": {"count": {"_gt": 1}}}}, {"components": {"code": {"_eq": "test"}}}]}
    )

    inner_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .clauses[0]
        .right.element._having_criteria[0]
        .clauses[1]
    )
    assert inner_having_stmt.left.name == "count"
    assert inner_having_stmt.right.value == 1
    assert inner_having_stmt.operator == operator.gt

    component_code_stmt = (
        db_query._where_criteria[0].right.element._raw_columns[0].element._where_criteria[0].clauses[1].clauses[1]
    )
    assert component_code_stmt.left.key == "code"
    assert component_code_stmt.operator == operator.eq

    systems = db_query.all()
    assert len(systems) == 1
    assert systems[0].uuid == system.uuid


def test_having_count_with_combined_filter(query, system, factories):
    other_system = factories.System()
    factories.Component(system=other_system, code="test")

    factories.Component(system=system, code="test")
    factories.Component(system=system)
    factories.Component(system=system)
    db_query = query({"components": {"_having": {"count": {"_gt": 1}}, "code": {"_eq": "test"}}})

    inner_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .right.element._having_criteria[0]
        .clauses[1]
    )
    assert inner_having_stmt.left.name == "count"
    assert inner_having_stmt.right.value == 1
    assert inner_having_stmt.operator == operator.gt

    inner_code_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .right.element._where_criteria[0]
        .clauses[1]
    )
    assert inner_code_stmt.left.key == "code"
    assert inner_code_stmt.operator == operator.eq

    assert db_query.count() == 0


def test_having_count_with_or_filter(query, system, factories):
    other_system = factories.System()
    factories.Component(system=other_system, code="test")

    factories.Component(system=system, code="test")
    factories.Component(system=system)
    factories.Component(system=system)
    db_query = query(
        {"_or": [{"components": {"_having": {"count": {"_gt": 1}}}}, {"components": {"code": {"_eq": "test"}}}]}
    )

    inner_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .clauses[0]
        .right.element._having_criteria[0]
        .clauses[1]
    )
    assert inner_having_stmt.left.name == "count"
    assert inner_having_stmt.right.value == 1
    assert inner_having_stmt.operator == operator.gt

    inner_code_stmt = (
        db_query._where_criteria[0].right.element._raw_columns[0].element._where_criteria[0].clauses[1].clauses[1]
    )
    assert inner_code_stmt.left.key == "code"
    assert inner_code_stmt.operator == operator.eq

    assert db_query.count() == 2


def test_having_sum_datapoint_float_return_nothing(session, factories):
    factories.DatapointFloat.create_batch(2)
    query = session.query(model.Component)
    configuration = query_configurations.Component(model.Component)
    query_filter = {"float_datapoints": {"_having": {"count": {"_gte": 1}, "sum": {"_gt": 300}}}}
    db_query = configuration.apply(query, query_filter)

    inner_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .right.element._having_criteria[0]
    )

    assert len(inner_having_stmt.clauses) == 2

    having_count_stmt = inner_having_stmt.clauses[0].clauses[1]
    assert having_count_stmt.left.name == "count"
    assert having_count_stmt.right.value == 1
    assert having_count_stmt.operator == operator.ge

    having_sum_stmt = inner_having_stmt.clauses[1].clauses[1]
    assert having_sum_stmt.left.name == "sum"
    assert having_sum_stmt.right.value == 300
    assert having_sum_stmt.operator == operator.gt

    assert db_query.count() == 0


def test_having_sum_datapoint_float(session, factories):
    component = factories.Component()
    datapoints = factories.DatapointFloat.create_batch(2, component=component)
    factories.DatapointFloat.create_batch(20)
    query = session.query(model.Component)
    configuration = query_configurations.Component(model.Component)
    datapoint_sum = sum(datapoint.value for datapoint in datapoints)
    query_filter = {"float_datapoints": {"_having": {"count": {"_gt": 1}, "sum": {"_gte": datapoint_sum}}}}
    db_query = configuration.apply(query, query_filter)

    inner_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .right.element._having_criteria[0]
    )

    assert len(inner_having_stmt.clauses) == 2

    having_count_stmt = inner_having_stmt.clauses[0].clauses[1]
    assert having_count_stmt.left.name == "count"
    assert having_count_stmt.right.value == 1
    assert having_count_stmt.operator == operator.gt

    having_sum_stmt = inner_having_stmt.clauses[1].clauses[1]
    assert having_sum_stmt.left.name == "sum"
    assert having_sum_stmt.right.value == datapoint_sum
    assert having_sum_stmt.operator == operator.ge

    components = db_query.all()
    assert len(components) == 1
    assert components[0].uuid == component.uuid


def test_having_sum_datapoint_float_component_code(session, factories):
    component = factories.Component()
    datapoints = factories.DatapointFloat.create_batch(2, component=component)
    factories.DatapointFloat.create_batch(20)
    query = session.query(model.Component)
    configuration = query_configurations.Component(model.Component)
    datapoint_sum = sum(datapoint.value for datapoint in datapoints)
    query_filter = {
        "code": {"_eq": component.code},
        "float_datapoints": {"_having": {"count": {"_gt": 1}, "sum": {"_gte": datapoint_sum}}},
    }
    db_query = configuration.apply(query, query_filter)

    code_where_filter = (
        db_query._where_criteria[0].right.element._raw_columns[0].element._where_criteria[0].clauses[0].clauses[1]
    )

    assert code_where_filter.left.key == "code"
    assert code_where_filter.operator == operator.eq

    inner_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .clauses[1]
        .right.element._having_criteria[0]
    )

    assert len(inner_having_stmt.clauses) == 2

    having_count_stmt = inner_having_stmt.clauses[0].clauses[1]
    assert having_count_stmt.left.name == "count"
    assert having_count_stmt.right.value == 1
    assert having_count_stmt.operator == operator.gt

    having_sum_stmt = inner_having_stmt.clauses[1].clauses[1]
    assert having_sum_stmt.left.name == "sum"
    assert having_sum_stmt.right.value == datapoint_sum
    assert having_sum_stmt.operator == operator.ge

    components = db_query.all()
    assert len(components) == 1
    assert components[0].uuid == component.uuid


def test_having_sum_datapoints_one_component_count(session, factories):
    component = factories.Component()
    factories.DatapointFloat.create_batch(2, value=3.0, component=component)
    factories.DatapointFloat.create_batch(20, value=-2.0)
    query = session.query(model.DatapointFloat.component_uuid).group_by(model.DatapointFloat.component_uuid)
    configuration = query_configurations.DatapointFloat(model.DatapointFloat)
    query_filter = {"_having": {"sum": {"_gt": 5.0}}, "component": {"_having": {"count": {"_eq": 1}}}}
    db_query = configuration.apply(query, query_filter)

    root_having_stmt = db_query._having_criteria[0].clauses[1]
    assert root_having_stmt.left.name == "sum"
    assert root_having_stmt.right.value == 5.0
    assert root_having_stmt.operator == operator.gt

    inner_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .right.element._having_criteria[0]
        .clauses[1]
    )
    assert inner_having_stmt.left.name == "count"
    assert inner_having_stmt.right.value == 1
    assert inner_having_stmt.operator == operator.eq

    assert db_query.count() == 1
    assert db_query.first().component_uuid == component.uuid


def test_having_sum_datapoints_invalid_component_count(session, factories):
    component = factories.Component()
    factories.DatapointFloat.create_batch(2, value=3.0, component=component)
    factories.DatapointFloat.create_batch(20, value=-2.0)
    query = session.query(model.DatapointFloat.component_uuid).group_by(model.DatapointFloat.component_uuid)
    configuration = query_configurations.DatapointFloat(model.DatapointFloat)
    query_filter = {"_having": {"sum": {"_gt": 5.0}}, "component": {"_having": {"count": {"_gt": 1}}}}
    db_query = configuration.apply(query, query_filter)

    root_having_stmt = db_query._having_criteria[0].clauses[1]
    assert root_having_stmt.left.name == "sum"
    assert root_having_stmt.right.value == 5.0
    assert root_having_stmt.operator == operator.gt

    inner_having_stmt = (
        db_query._where_criteria[0]
        .right.element._raw_columns[0]
        .element._where_criteria[0]
        .right.element._having_criteria[0]
        .clauses[1]
    )
    assert inner_having_stmt.left.name == "count"
    assert inner_having_stmt.right.value == 1
    assert inner_having_stmt.operator == operator.gt

    assert db_query.count() == 0


def test_nested_having_groups_on_datapoints(session, factories):
    component = factories.Component()
    factories.DatapointFloat.create_batch(2, value=3.0, component=component)
    factories.DatapointFloat.create_batch(20, value=-2.0)
    query = session.query(model.DatapointFloat.component_uuid).group_by(model.DatapointFloat.component_uuid)
    configuration = query_configurations.DatapointFloat(model.DatapointFloat)
    query_filter = {"_having": {"sum": {"_gt": 5.0}, "nested_having_group": {"avg": {"_gte": 3.0}}}}
    db_query = configuration.apply(query, query_filter)

    assert db_query.count() == 1
    assert db_query.first().component_uuid == component.uuid


def test_having_root_object_no_where_statement(session, factories):
    component = factories.Component()
    factories.DatapointFloat.create_batch(2, value=3.0, component=component)
    configuration = query_configurations.DatapointFloat(model.DatapointFloat)
    query_filter = {"_having": {"sum": {"_gt": 5.0}}}
    query = session.query(model.DatapointFloat).group_by(model.DatapointFloat.component_uuid)
    query = configuration.apply(query, query_filter)

    assert len(query._where_criteria) == 0
