from typing import Any, Dict

import pytest
from sqlalchemy.orm.query import Query, aliased

from xcnt.sqlalchemy.search import EvaluationErrorCode
from xcnt.sqlalchemy.search.test import model
from xcnt.sqlalchemy.search.test import query as query_configurations


@pytest.fixture
def system(factories):
    return factories.System()


@pytest.fixture
def systems(factories):
    return factories.System.create_batch(20)


@pytest.fixture
def query_object():
    return model.System


@pytest.fixture
def query(session, query_object):
    def _query(query_filter: Dict[str, Any]) -> Query:
        query = session.query(query_object)
        configuration = query_configurations.System(query_object)
        return configuration.apply(query, query_filter)

    return _query


@pytest.fixture
def result_for(session, query_object):
    def _result_for(result_filter: Dict[str, Any]) -> Query:
        configuration = query_configurations.System(query_object)
        return configuration.generate(result_filter)

    return _result_for


@pytest.mark.parametrize("query_object", [model.System, aliased(model.System)])
def test_neq_with_null_values(query, system, factories):
    second_system = factories.System()
    factories.System(alias=None)

    db_query = query({"alias": {"_neq": system.alias, "_null": True}})
    result = db_query.all()
    assert len(result) == 2
    for system_result in result:
        assert system_result.alias in [second_system.alias, None]


@pytest.mark.parametrize("query_object", [model.System, aliased(model.System)])
def test_neq_without_null_values(query, system, factories):
    second_system = factories.System()
    factories.System.create(alias=None)

    db_query = query({"alias": {"_neq": system.alias, "_null": False}})
    result = db_query.all()
    assert len(result) == 1
    assert result[0].alias == second_system.alias


@pytest.mark.parametrize("query_object", [model.System, aliased(model.System)])
def test_eq_return_null_values(query, system, factories):
    factories.System(alias=None)

    db_query = query({"alias": {"_eq": None, "_null": True}})
    assert db_query.count() == 1


@pytest.mark.parametrize("query_object", [model.System, aliased(model.System)])
def test_eq_do_not_return_null_values(query, system, factories):
    factories.System(alias=None)

    db_query = query({"alias": {"_eq": None, "_null": False}})
    assert db_query.count() == 0


@pytest.mark.parametrize("key", ["_in", "_nin"])
def test_list_without_null_values(query, key, factories, systems):
    factories.System(alias=None)
    query_object = query({"alias": {key: None, "_null": False}})
    assert query_object.count() == 20


@pytest.mark.parametrize("key", ["_in", "_nin"])
def test_list_with_null_values(query, key, factories, systems):
    factories.System(alias=None)
    query_object = query({"alias": {key: None, "_null": True}})
    assert query_object.count() == 21


def test_invalid_null_value_argument(result_for):
    result = result_for({"alias": {"_neq": "123", "_null": "invalid_argument"}})
    assert result.errors["alias"]["_neq"]["_null"] == EvaluationErrorCode.INVALID_VALUE_FOR_FILTER


def test_invalid_null_invalid_argument(result_for):
    result = result_for({"alias": {"_null": "invalid_argument"}})
    assert result.errors["alias"]["_null"] == EvaluationErrorCode.INVALID_VALUE_FOR_FILTER


def test_invalid_null_return_all_values_with_none(query, factories, systems):
    factories.System(alias=None)
    query_object = query({"alias": {"_null": True}})
    assert query_object.count() == 1


def test_invalid_null_return_all_values_without_none(query, factories, systems):
    factories.System(alias=None)
    query_object = query({"alias": {"_null": False}})
    assert query_object.count() == 20
